#!/usr/bin/env python3
# encoding: utf-8

from __future__ import annotations

__all__ = ["P115ZipPath", "P115ZipFileSystem"]

from collections.abc import (
    AsyncIterator, Coroutine, Iterable, Iterator, MutableMapping, 
)
from itertools import count
from os import PathLike
from posixpath import basename, dirname
from threading import Lock
from typing import overload, Any, Literal

from dictattr import AttrDict
from dicttools import dict_key_to_lower_update
from errno2 import errno
from iterutils import run_gen_step, run_gen_step_iter, with_iter_next, Yield

from ..client import check_response, P115Client
from ..exception import throw
from ..tool import extract_iter_files, extract_iterdir
from ..type import P115URL
from ..util import lock_as_async
from .fs_base import IDOrPathType, P115PathBase, P115FileSystemBase


# TODO: 尽量也要兼容 zipfile.Path 的接口
class P115ZipPath(P115PathBase):
    __slots__ = ("fs", "attr")
    fs: P115ZipFileSystem


class P115ZipFileSystem(P115FileSystemBase[P115ZipPath]):
    path_class = P115ZipPath

    def __init__(
        self, 
        /, 
        client: str | PathLike | P115Client, 
        pickcode: str, 
        refresh: bool = False, 
        id_to_readdir: None | int | dict[int, dict[int, MutableMapping]] = None, 
    ):
        super().__init__(client, refresh=refresh, id_to_readdir=id_to_readdir)
        self.pickcode = pickcode
        self._path_to_id: dict[str, int] = {"/": 0}
        self._id_to_path: dict[int, str] = {0: "/"}
        self._load_lock = Lock()

    @overload
    def id_to_path(
        self, 
        id: int, 
        /, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> str:
        ...
    @overload
    def id_to_path(
        self, 
        id: int, 
        /, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, str]:
        ...
    def id_to_path(
        self, 
        id: int, 
        /, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> str | Coroutine[Any, Any, str]:
        def gen_step():
            if not self.full_loaded:
                yield self.load(locked=True, async_=async_, **request_kwargs)
            try:
                return self._id_to_path[id]
            except KeyError:
                throw(errno.ENOENT, id)
        return run_gen_step(gen_step, async_)

    @overload
    def path_to_id(
        self, 
        path: str, 
        /, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> int:
        ...
    @overload
    def path_to_id(
        self, 
        path: str, 
        /, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, int]:
        ...
    def path_to_id(
        self, 
        path: str, 
        /, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> int | Coroutine[Any, Any, int]:
        def gen_step():
            if not self.full_loaded:
                yield self.load(locked=True, async_=async_, **request_kwargs)
            try:
                return self._path_to_id[path]
            except KeyError:
                throw(errno.ENOENT, path)
        return run_gen_step(gen_step, async_)

    @overload
    def extract(
        self, 
        /, 
        path: IDOrPathType = "", 
        to_pid: int | str = 0, 
        pid: None | int = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> int:
        ...
    @overload
    def extract(
        self, 
        /, 
        path: IDOrPathType = "", 
        to_pid: int | str = 0, 
        pid: None | int = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, int]:
        ...
    def extract(
        self, 
        /, 
        path: IDOrPathType = "", 
        to_pid: int | str = 0, 
        pid: None | int = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> int | Coroutine[Any, Any, int]:
        """解压缩到网盘
        """
        def gen_step():
            payload = [
                ("pick_code", self.pickcode), 
                ("paths", "文件"), 
                ("to_pid", to_pid), 
            ]
            add_file = payload.append
            attr: MutableMapping = self.get_attr(path, pid=pid)
            if attr["id"]:
                add_file((
                    "extract_dir" if attr["is_dir"] else "extract_file", 
                    attr["path"].strip("/"), 
                ))
            else:
                for attr in self.readdir(0):
                    add_file((
                        "extract_dir[]" if attr["is_dir"] else "extract_file[]", 
                        attr["path"].strip("/"), 
                    ))
            resp = yield self.client.extract_add_file_app(
                payload, 
                async_=async_, 
                **request_kwargs, 
            )
            check_response(resp)
            return resp["data"]["extract_id"]
        return run_gen_step(gen_step, async_)

    @overload
    def extract_many(
        self, 
        /, 
        paths: Iterable[IDOrPathType], 
        to_pid: int | str = 0, 
        pid: None | int = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> int:
        ...
    @overload
    def extract_many(
        self, 
        /, 
        paths: Iterable[IDOrPathType], 
        to_pid: int | str = 0, 
        pid: None | int = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, int]:
        ...
    def extract_many(
        self, 
        /, 
        paths: Iterable[IDOrPathType], 
        to_pid: int | str = 0, 
        pid: None | int = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> int | Coroutine[Any, Any, int]:
        """解压缩到网盘
        """
        def gen_step():
            payload = [
                ("pick_code", self.pickcode), 
                ("paths", "文件"), 
                ("to_pid", to_pid), 
            ]
            add_file = payload.append
            for p in paths:
                attr = self.get_attr(p, pid=pid)
                add_file((
                    "extract_dir[]" if attr["is_dir"] else "extract_file[]", 
                    attr["path"].strip("/"), 
                ))
            resp = yield self.client.extract_add_file_app(
                payload, 
                async_=async_, 
                **request_kwargs, 
            )
            check_response(resp)
            return resp["data"]["extract_id"]
        return run_gen_step(gen_step, async_)

    @overload
    def get_url(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> P115URL:
        ...
    @overload
    def get_url(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, P115URL]:
        ...
    def get_url(
        self, 
        id_or_path: IDOrPathType, 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> P115URL | Coroutine[Any, Any, P115URL]:
        "获取下载链接"
        def gen_step():
            headers = request_kwargs["headers"] = dict(request_kwargs.get("headers") or ())
            dict_key_to_lower_update(headers)
            user_agent = headers.pop("user-agent", "")
            attr = yield self.get_attr(
                id_or_path, 
                pid=pid, 
                ensure_file=True, 
                async_=async_, 
                **request_kwargs, 
            )
            return self.client.extract_download_url(
                self.pickcode, 
                attr["path"], 
                user_agent=user_agent, 
                async_=async_, 
                **request_kwargs, 
            )
        return run_gen_step(gen_step, async_)

    @overload
    def iterdir(
        self, 
        id: int, 
        /, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> Iterator[dict]:
        ...
    @overload
    def iterdir(
        self, 
        id: int, 
        /, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> AsyncIterator[dict]:
        ...
    def iterdir(
        self, 
        id: int, 
        /, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> Iterator[dict] | AsyncIterator[dict]:
        """迭代获取目录内直属的文件或目录的信息
        """
        def gen_step():
            if not self.full_loaded:
                yield self.load(locked=True, async_=async_, **request_kwargs)
            if id:
                try:
                    attr = self.id_to_attr[id]
                except KeyError:
                    throw(errno.ENOENT, id)
                if not attr["is_dir"]:
                    throw(errno.ENOTDIR, attr)
                path = attr["path"]
            else:
                path = "/"
            with with_iter_next(extract_iterdir(
                self.client, 
                self.pickcode, 
                path, 
                async_=async_, 
                **request_kwargs, 
            )) as get_next:
                while True:
                    attr = yield get_next()
                    try:
                        attr["id"] = self._path_to_id[attr["path"]]
                    except KeyError:
                        continue
                    yield Yield(attr)
        return run_gen_step_iter(gen_step, async_)

    @overload
    def readdir(
        self, 
        id_or_path: IDOrPathType = "", 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> list[MutableMapping]:
        ...
    @overload
    def readdir(
        self, 
        id_or_path: IDOrPathType = "", 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, list[MutableMapping]]:
        ...
    def readdir(
        self, 
        id_or_path: IDOrPathType = "", 
        /, 
        pid: None | int = None, 
        refresh: None | bool = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> list[MutableMapping] | Coroutine[Any, Any, list[MutableMapping]]:
        readdir = super().readdir
        def gen_step():
            if not self.full_loaded:
                yield self.load(locked=True, async_=async_, **request_kwargs)
            return readdir(
                id_or_path, 
                pid=pid, 
                refresh=refresh, 
                async_=async_, 
                **request_kwargs, 
            )
        return run_gen_step(gen_step, async_)

    @overload
    def load(
        self, 
        /, 
        locked: bool = True, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> None:
        ...
    @overload
    def load(
        self, 
        /, 
        locked: bool = False, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, None]:
        ...
    def load(
        self, 
        /, 
        locked: bool = False, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> None | Coroutine[Any, Any, None]:
        """一次性加载整个压缩包中的文件列表
        """
        def gen_step():
            if locked:
                if async_:
                    async def request():
                        async with lock_as_async(self._load_lock):
                            if not self.full_loaded:
                                await self.load(async_=True, **request_kwargs)
                    yield request()
                else:
                    with self._load_lock:
                        if not self.full_loaded:
                            self.load(**request_kwargs)
                return
            path_to_id    = self._path_to_id
            id_to_path    = self._id_to_path
            id_to_dirnode = self.id_to_dirnode
            id_to_readdir = self.id_to_readdir
            id_to_attr    = self.id_to_attr
            get_id = count(1).__next__
            id_to_readdir[0] = {}
            def get_parent_id(path: str, /) -> int:
                dir_ = dirname(path)
                if dir_ == "/":
                    return 0
                try:
                    return path_to_id[dir_]
                except KeyError:
                    pid = get_parent_id(dir_)
                    cid = path_to_id[dir_] = get_id()
                    id_to_path[cid] = dir_
                    name = basename(dir_)
                    id_to_readdir[pid][cid] = id_to_attr[cid] = AttrDict({
                        "id": cid, 
                        "parent_id": pid, 
                        "name": name, 
                        "is_dir": True, 
                        "path": dir_, 
                    })
                    id_to_dirnode[cid] = (name, pid)
                    id_to_readdir[cid] = {}
                    return cid
            with with_iter_next(extract_iter_files(
                self.client, 
                self.pickcode, 
                async_=async_, 
                **request_kwargs, 
            )) as get_next:
                while True:
                    attr: MutableMapping = AttrDict((yield get_next()))
                    attr["parent_id"] = pid = get_parent_id(attr["path"])
                    attr["id"] = id = get_id()
                    attr["is_dir"] = False
                    id_to_readdir[pid][id]= id_to_attr[id] = attr
                    path = attr["path"]
                    id_to_path[id] = path
                    path_to_id[attr["path"]] = id
            self.full_loaded = True
        return run_gen_step(gen_step, async_)

# TODO: 增加接口，可用于检查文件是否已经云解压
# resp = check_response(client.extract_push_progress(pickcode))
# if resp["data"]["extract_status"]["unzip_status"] != 4:
#     raise OSError(errno.EIO, "file was not decompressed")
# TODO: 增加接口，可用于推送云解压，云解压进度，解压到文件夹进度等
# TODO: 参考 zipfile 模块的接口设计 namelist、filelist 等属性，以及其它的和 zipfile 兼容的接口
# TODO: 当文件特别多时，可以用 zipfile 等模块来读取文件列表



#!/usr/bin/env python3
# encoding: utf-8

__all__ = [
    "P115Error", "P115Warning", "P115OSError", "P115AccessError", 
    "P115AuthenticationError", "P115BusyOSError", "P115DataError", 
    "P115OperationalError", "P115FileTooBig", "P115ExceededError", 
    "P115InvalidArgumentError", "P115NoSpaceError", "P115NotSupportedError", 
    "P115LoginError", "P115AccessTokenError", "P115OpenAppAuthLimitExceeded", 
    "error", "throw", "errno2error"
]

import warnings

from itertools import count
from collections.abc import Mapping
from functools import cached_property
from typing import Never

from errno2 import errno, errno2error as _errno2error


warnings.filterwarnings("always", category=UserWarning)
setattr(warnings, "formatwarning", lambda message, category, filename, lineno, line=None, _getid=count(1).__next__:
    f"\r\x1b[K\x1b[1;31;43m{category.__qualname__}\x1b[0m(\x1b[32m{_getid()}\x1b[0m) @ \x1b[3;4;34m{filename}\x1b[0m:\x1b[36m{lineno}\x1b[0m \x1b[5;31m➜\x1b[0m \x1b[1m{message}\x1b[0m\n"
)


class P115Error(Exception):
    """本模块的最基础异常类
    """


class P115Warning(P115Error, UserWarning):
    """本模块的最基础警示类
    """


class P115OSError(P115Error, OSError):
    """本模块的最基础输出输出异常类
    """
    def __getattr__(self, attr, /):
        try:
            return self[attr]
        except KeyError as e:
            raise AttributeError(attr) from e

    def __getitem__(self, key, /):
        message = self.message
        if isinstance(message, Mapping):
            return message[key]
        raise KeyError(key)

    @cached_property
    def message(self, /):
        if args := self.args:
            if len(args) >= 2 and isinstance(args[0], int):
                return args[1]
            return args[0]


class P115AuthenticationError(P115OSError):
    """当登录状态无效时抛出
    """


class P115BusyOSError(P115OSError):
    """当操作繁忙时抛出（115 网盘的复制、移动、删除、还原只允许最多一个操作进行中）
    """


class P115DataError(P115OSError):
    """当响应数据解析失败时抛出
    """


class P115OperationalError(P115OSError):
    """当接口使用方法错误时抛出，例如参数错误、空间不足、超出允许数量范围等
    """


class P115AccessError(P115OperationalError, PermissionError):
    """当不可访问时抛出，可能是文件被和谐了
    """


class P115FileTooBig(P115OperationalError, PermissionError):
    """文件过大
    """


class P115ExceededError(P115OperationalError, PermissionError):
    """超出允许数量范围
    """


class P115InvalidArgumentError(P115OperationalError, ValueError):
    """参数错误
    """


class P115NoSpaceError(P115OperationalError, PermissionError):
    """空间不足
    """


class P115NotSupportedError(P115OperationalError):
    """当调用不存在的接口或者接口不支持此操作时抛出
    """


class P115LoginError(P115AuthenticationError):
    """当登录失败时抛出
    """


class P115AccessTokenError(P115OSError, ValueError):
    """access_token 错误或者无效
    """


class P115OpenAppAuthLimitExceeded(P115AuthenticationError):
    """当授权应用数达到上限时抛出
    """


def error(*args, **kwds) -> BaseException:
    """构建异常

    .. tip::
        会根据传入的位置参数，做一些类型推断

        - 第 1 个位置参数，记作 `errcode`，大概是一个 `errno2.errno` 的枚举类型，不能成功推断则用 `errno2.errno.EIO`
        - 第 2 个位置参数（若第 1 个位置参数不满足上一条，则用此参数），记作 `exctype`，大概是一个 `P115Error` 类型或其子类型，不能成功推断则用 `P115OSError`

        假设剩余的所有没被提取的位置参数记作 `rargs`，最终构建的异常为 `exctype(errcode, *rargs, **kwds)`
    """
    if args and isinstance(args[0], errno):
        errcode = args[0]
        args = args[1:]
    else:
        errcode = errno.EIO
    if not (args and isinstance(args[0], type) and issubclass(args[0], BaseException)):
        args = errno2error.get(errcode, P115OSError), *args
    return errcode.error(*args, **kwds)


def throw(*args, **kwds) -> Never:
    """抛出异常

    .. tip::
        会根据传入的位置参数，做一些类型推断

        - 第 1 个位置参数，记作 `errcode`，大概是一个 `errno2.errno` 的枚举类型，不能成功推断则用 `errno2.errno.EIO`
        - 第 2 个位置参数（若第 1 个位置参数不满足上一条，则用此参数），记作 `exctype`，大概是一个 `P115Error` 类型或其子类型，不能成功推断则用 `P115OSError`

        假设剩余的所有没被提取的位置参数记作 `rargs`，最终抛出的异常为 `exctype(errcode, *rargs, **kwds)`
    """
    raise error(*args, **kwds)


#: errno 到的异常类的映射
errno2error: dict[errno, type[P115Error]] = {
    errno.EAUTH: P115AuthenticationError, 
    errno.EBUSY: P115BusyOSError, 
    errno.ENODATA: P115DataError, 
    errno.EACCES: P115AccessError, 
    errno.EFBIG: P115FileTooBig, 
    errno.ERANGE: P115ExceededError, 
    errno.EINVAL: P115InvalidArgumentError, 
    errno.ENOSPC: P115NoSpaceError, 
    errno.ENOTSUP: P115NotSupportedError, 
    errno.ENOSYS: P115NotSupportedError, 
}

_modns = globals()
for errcode, exctype in _errno2error.items():
    typename = "P115" + exctype.__name__
    if typename not in _modns:
        cls = _modns[typename] = type(typename, (P115OSError, exctype), {"__doc__": errcode.description})
        __all__.append(typename)
        errno2error[errcode] = cls


def __getattr__(attr: str, /) -> type[P115OSError]:
    raise AttributeError(attr)




#!/usr/bin/env python3
# encoding: utf-8

from __future__ import annotations

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = [
    "AVAILABLE_APPS", "APP_TO_SSOENT", "SSOENT_TO_APP", "CLIENT_METHOD_API_MAP", 
    "CLIENT_API_METHODS_MAP", "CLASS_TO_TYPE", "SUFFIX_TO_TYPE", "TYPE_TO_SUFFIXES", 
    "ID_TO_DIRNODE_CACHE", "WEPAPI_PREFIXES", 
]

from collections import defaultdict
from collections.abc import MutableMapping
from pathlib import Path
from typing import Final

from iter_collect import grouped_mapping


_CACHE_DIR = Path("~/.p115client.cache.d").expanduser()
_CACHE_DIR.mkdir(exist_ok=True)

#: 目前可用的登录设备
AVAILABLE_APPS: Final[dict[str, str]] = {
    "web": "115生活_网页端", 
    "ios": "115生活_苹果端", 
    "115ios": "115_苹果端", 
    "android": "115生活_安卓端", 
    "115android": "115_安卓端", 
    "ipad": "115生活_苹果平板端", 
    "115ipad": "115_苹果平板端", 
    "tv": "115生活_安卓电视端", 
    "apple_tv": "115生活_苹果电视端", 
    "qandroid": "115管理_安卓端", 
    "qios": "115管理_苹果端", 
    "qipad": "115管理_苹果平板端", 
    # "windows": "115生活_Windows端", 
    # "mac": "115生活_macOS端", 
    # "linux": "115生活_Linux端", 
    "wechatmini": "115生活_微信小程序端", 
    "alipaymini": "115生活_支付宝小程序", 
    "harmony": "115_鸿蒙端", 
}

#: 目前已知的登录设备和对应的 ssoent
APP_TO_SSOENT: Final[dict[str, str]] = {
    "web": "A1", 
    "desktop": "A1", 
    "ios": "D1", 
    "bios": "D2", 
    "115ios": "D3", 
    "android": "F1", 
    "bandroid": "F2", 
    "115android": "F3", 
    "ipad": "H1", 
    "bipad": "H2", 
    "115ipad": "H3", 
    "tv": "I1", 
    "apple_tv": "I2", 
    "qandroid": "M1", 
    "qios": "N1", 
    "qipad": "O1", 
    "windows": "P1", 
    "mac": "P2", 
    "linux": "P3", 
    "wechatmini": "R1", 
    "alipaymini": "R2", 
    "harmony": "S1", 
}

#: 目前已知的 ssoent 和对应的登录设备，一部分因为不知道具体的设备名，所以使用目前可用的设备名，作为临时代替
SSOENT_TO_APP: Final[dict[str, str]] = {v: k for k, v in reversed(APP_TO_SSOENT.items())}

#: 所有已封装的方法名和对应的 115 接口
CLIENT_METHOD_API_MAP: Final[dict[str, str]] = {}

#: 所有已封装的 115 接口和对应的方法名
CLIENT_API_METHODS_MAP: Final[dict[str, list[str]]] = {}

#: 文件的 class 属性对应的所属类型的整数代码
CLASS_TO_TYPE: Final[dict[str, int]] = {
    "JG_DOC": 1, 
    "DOC": 1, 
    "JG_PIC": 2, 
    "PIC": 2, 
    "JG_MUS": 3, 
    "MUS": 3, 
    "JG_AVI": 4, 
    "AVI": 4, 
    "JG_RAR": 5, 
    "RAR": 5, 
    "RAR_EXTRACT": 5, 
    "JG_EXE": 6, 
    "EXE": 6, 
    "JG_BOOK": 7, 
    "BOOK": 7
}

#: 文件后缀对应的所属类型的整数代码（尚需补充）
SUFFIX_TO_TYPE: Final[dict[str, int]] = {
    ".ass": 1, 
    ".chm": 1, 
    ".doc": 1, 
    ".docm": 1, 
    ".docx": 1, 
    ".dotm": 1, 
    ".dot": 1, 
    ".dwg": 1, 
    ".htm": 1, 
    ".html": 1, 
    ".idx": 1, 
    ".jar": 1, 
    ".key": 1, 
    ".log": 1, 
    ".lrc": 1, 
    ".mdb": 1, 
    ".mdf": 1, 
    ".numbers": 1, 
    ".ods": 1, 
    ".odt": 1, 
    ".pages": 1, 
    ".pdf": 1, 
    ".pot": 1, 
    ".pps": 1, 
    ".ppt": 1, 
    ".pptm": 1, 
    ".pptx": 1, 
    ".rtf": 1, 
    ".srt": 1, 
    ".ssa": 1, 
    ".sub": 1, 
    ".torrent": 1, 
    ".txt": 1, 
    ".wps": 1, 
    ".wri": 1, 
    ".xlam": 1, 
    ".xls": 1, 
    ".xlsb": 1, 
    ".xlsm": 1, 
    ".xlsx": 1, 
    ".xltm": 1, 
    ".xltx": 1, 
    ".vtt": 1, 
    ".bmp": 2, 
    ".exif": 2, 
    ".gif": 2, 
    ".heic": 2, 
    ".heif": 2, 
    ".jpeg": 2, 
    ".jpg": 2, 
    ".png": 2, 
    ".raw": 2, 
    ".svg": 2, 
    ".tif": 2, 
    ".tiff": 2, 
    ".webp": 2, 
    ".aac": 3, 
    ".aiff": 3, 
    ".amr": 3, 
    ".ape": 3, 
    ".au": 3, 
    ".f4a": 3, 
    ".flac": 3, 
    ".m4a": 3, 
    ".mid": 3, 
    ".midi": 3, 
    ".mp3": 3, 
    ".ogg": 3, 
    ".wav": 3, 
    ".wma": 3, 
    ".3g2": 4, 
    ".3gp": 4, 
    ".3gp2": 4, 
    ".3gpp": 4, 
    ".asf": 4, 
    ".avi": 4, 
    ".dat": 4, 
    ".divx": 4, 
    ".f4v": 4, 
    ".flv": 4, 
    ".iso": 4, 
    ".m2ts": 4, 
    ".m4v": 4, 
    ".mkv": 4, 
    ".mov": 4, 
    ".mp4": 4, 
    ".mpe": 4, 
    ".mpeg": 4, 
    ".mpeg4": 4, 
    ".mpg": 4, 
    ".mts": 4, 
    ".ram": 4, 
    ".rm": 4, 
    ".rmvb": 4, 
    ".swf": 4, 
    ".ts": 4, 
    ".vob": 4, 
    ".webm": 4, 
    ".wmv": 4, 
    ".7z": 5, 
    ".cab": 5, 
    ".dmg": 5, 
    ".msi": 5, 
    ".rar": 5, 
    ".tar": 5, 
    ".xz": 5, 
    ".z": 5, 
    ".zip": 5, 
    ".apk": 6, 
    ".bat": 6, 
    ".deb": 6, 
    ".exe": 6, 
    ".ipa": 6, 
    ".pkg": 6, 
    ".azw": 7, 
    ".azw3": 7, 
    ".epub": 7, 
    ".fb2": 7, 
    ".lit": 7, 
    ".lrf": 7, 
    ".mobi": 7, 
    ".prc": 7, 
}

#: 文件类型的整数代码对应的扩展名元组
TYPE_TO_SUFFIXES: Final[dict[int, tuple[str, ...]]] = {
    k: tuple(v) for k, v in grouped_mapping(
        (v, k) for k, v in SUFFIX_TO_TYPE.items()).items()}

#: 用于缓存每个用户或者分享，的每个目录 id 到所对应的 (name, parent_id) 的元组的字典的字典
ID_TO_DIRNODE_CACHE: Final[dict[int | str, MutableMapping[int, tuple[str, int]]]] = defaultdict(dict)

#: 一些 webapi.115.com 的接口允许的前缀（而 proapi.115.com 却可以是任何前缀）
WEPAPI_PREFIXES = (
    "/behavior", "/category", "/files", "/history", "/label", "/movies", 
    "/offine", "/photo", "/rb", "/share", "/user", "/usershare", 
)



#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"

from p115client.util import *

from .attr import *
from .auth import *
from .download import *
from .edit import *
from .export_dir import *
from .fs_files import *
from .history import *
from .iterdir import *
from .life import *
from .offline import *
from .pool import *
from .updatedb import *
from .upload import *
from .xys import *



#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = [
    "update_abstract", "update_desc", "update_star", "update_label", "update_score", 
    "update_top", "update_show_play_long", "update_category_shortcut", "batch_unstar", 
    "update_name", "post_event", "batch_makedir", "copyfile", "renamefile", "transferfile", 
]
__doc__ = "这个模块提供了一些和修改文件或目录信息有关的函数"

from asyncio import Future as AsyncFuture
from collections.abc import (
    AsyncIterable, AsyncIterator, Callable, Coroutine, Iterable, 
    Iterator, Mapping, 
)
from concurrent.futures import Future
from functools import partial
from itertools import batched
from os import PathLike
from typing import overload, Any, Literal

from concurrenttools import conmap, run_as_thread, run_as_async
from iterutils import chunked, map as do_map, run_gen_step, as_gen_step, through
from p115pickcode import to_id

from ..client import check_response, P115Client, P115OpenClient
from ..type import P115URL


@overload
def update_abstract(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str], 
    /, 
    method: str, 
    value: Any, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
):
    ...
@overload
def update_abstract(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    method: str, 
    value: Any, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine:
    ...
def update_abstract(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    method: str, 
    value: Any, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
):
    """批量设置文件或目录

    :param client: 115 客户端或 cookies
    :param ids: 一组文件或目录的 id 或 pickcode
    :param method: 方法名
    :param value: 要设置的值
    :param batch_size: 批次大小，分批次，每次提交的 id 数
    :param max_workers: 并发工作数，如果为 None 或者 <= 0，则自动确定
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    def gen_step():
        setter = partial(getattr(client, method), async_=async_, **request_kwargs)
        def call(batch, /):
            return check_response(setter(batch, value))
        yield through(conmap(
            call, 
            chunked(do_map(to_id, ids), batch_size), 
            max_workers=max_workers, 
            async_=async_, 
        ))
    return run_gen_step(gen_step, async_)


@overload
def update_desc(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str], 
    /, 
    desc: str = "", 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
):
    ...
@overload
def update_desc(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    desc: str = "", 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine:
    ...
def update_desc(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    desc: str = "", 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
):
    """批量给文件或目录设置备注，此举可更新此文件或目录的 mtime

    :param client: 115 客户端或 cookies
    :param ids: 一组文件或目录的 id 或 pickcode
    :param desc: 备注文本
    :param batch_size: 批次大小，分批次，每次提交的 id 数
    :param max_workers: 并发工作数，如果为 None 或者 <= 0，则自动确定
    :param app: 使用此设备的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数
    """
    if app in ("", "web", "desktop", "harmony"):
        method = "fs_desc_set"
    else:
        method = "fs_desc_set_app"
        request_kwargs["app"] = app
    return update_abstract(
        client, 
        ids, # type: ignore
        method=method, 
        value=desc, 
        batch_size=batch_size, 
        max_workers=max_workers, 
        async_=async_, # type: ignore
        **request_kwargs, 
    )


@overload
def update_star(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str], 
    /, 
    star: bool = True, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
):
    ...
@overload
def update_star(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    star: bool = True, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine:
    ...
def update_star(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    star: bool = True, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
):
    """批量给文件或目录设置星标

    .. note::
        如果一批中有任何一个 id 已经被删除，则这一批直接失败报错

    :param client: 115 客户端或 cookies
    :param ids: 一组文件或目录的 id 或 pickcode
    :param star: 是否设置星标
    :param batch_size: 批次大小，分批次，每次提交的 id 数
    :param max_workers: 并发工作数，如果为 None 或者 <= 0，则自动确定
    :param app: 使用此设备的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if not isinstance(client, P115Client) or app == "open":
        method = "fs_star_set_open"
    elif app in ("", "web", "desktop", "harmony"):
        method = "fs_star_set"
    else:
        method = "fs_star_set_app"
        request_kwargs["app"] = app
    return update_abstract(
        client, 
        ids, # type: ignore
        method=method, 
        value=star, 
        batch_size=batch_size, 
        max_workers=max_workers, 
        async_=async_, # type: ignore
        **request_kwargs, 
    )


@overload
def update_label(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str], 
    /, 
    label: int | str = 1, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
):
    ...
@overload
def update_label(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    label: int | str = 1, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine:
    ...
def update_label(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    label: int | str = 1, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
):
    """批量给文件或目录设置标签

    :param client: 115 客户端或 cookies
    :param ids: 一组文件或目录的 id 或 pickcode
    :param label: 标签 id，多个用逗号 "," 隔开，如果用一个根本不存在的 id，效果就是清空标签列表
    :param batch_size: 批次大小，分批次，每次提交的 id 数
    :param max_workers: 并发工作数，如果为 None 或者 <= 0，则自动确定
    :param app: 使用此设备的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数
    """
    if app in ("", "web", "desktop", "harmony"):
        method = "fs_label_set"
    else:
        method = "fs_label_set_app"
        request_kwargs["app"] = app
    return update_abstract(
        client, 
        ids, # type: ignore
        method=method, 
        value=label, 
        batch_size=batch_size, 
        max_workers=max_workers, 
        async_=async_, # type: ignore
        **request_kwargs, 
    )


@overload
def update_score(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str], 
    /, 
    score: int = 0, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
):
    ...
@overload
def update_score(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    score: int = 0, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine:
    ...
def update_score(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    score: int = 0, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
):
    """批量给文件或目录设置分数

    :param client: 115 客户端或 cookies
    :param ids: 一组文件或目录的 id 或 pickcode
    :param score: 分数
    :param batch_size: 批次大小，分批次，每次提交的 id 数
    :param max_workers: 并发工作数，如果为 None 或者 <= 0，则自动确定
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数
    """
    return update_abstract(
        client, 
        ids, # type: ignore
        method="fs_score_set", 
        value=score, 
        batch_size=batch_size, 
        max_workers=max_workers, 
        async_=async_, # type: ignore
        **request_kwargs, 
    )


@overload
def update_top(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str], 
    /, 
    top: bool = True, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
):
    ...
@overload
def update_top(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    top: bool = True, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine:
    ...
def update_top(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    top: bool = True, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
):
    """批量给文件或目录设置置顶

    :param client: 115 客户端或 cookies
    :param ids: 一组文件或目录的 id 或 pickcode
    :param score: 分数
    :param batch_size: 批次大小，分批次，每次提交的 id 数
    :param max_workers: 并发工作数，如果为 None 或者 <= 0，则自动确定
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数
    """
    return update_abstract(
        client, 
        ids, # type: ignore
        method="fs_top_set", 
        value=top, 
        batch_size=batch_size, 
        max_workers=max_workers, 
        async_=async_, # type: ignore
        **request_kwargs, 
    )


@overload
def update_show_play_long(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str], 
    /, 
    show: bool = True, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
):
    ...
@overload
def update_show_play_long(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    show: bool = True, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine:
    ...
def update_show_play_long(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    show: bool = True, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
):
    """批量给目录设置显示时长

    :param client: 115 客户端或 cookies
    :param ids: 一组目录的 id 或 pickcode
    :param show: 是否显示时长
    :param batch_size: 批次大小，分批次，每次提交的 id 数
    :param max_workers: 并发工作数，如果为 None 或者 <= 0，则自动确定
    :param app: 使用此设备的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数
    """
    if app in ("", "web", "desktop", "harmony"):
        method = "fs_show_play_long_set"
    else:
        method = "fs_show_play_long_set_app"
        request_kwargs["app"] = app
    return update_abstract(
        client, 
        ids, # type: ignore
        method=method, 
        value=show, 
        batch_size=batch_size, 
        max_workers=max_workers, 
        async_=async_, # type: ignore
        **request_kwargs, 
    )


@overload
def update_category_shortcut(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str], 
    /, 
    set: bool = True, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
):
    ...
@overload
def update_category_shortcut(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    set: bool = True, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine:
    ...
def update_category_shortcut(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    set: bool = True, 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
):
    """批量给目录设置显示时长

    :param client: 115 客户端或 cookies
    :param ids: 一组目录的 id 或 pickcode
    :param set: 是否设为快捷入口
    :param batch_size: 批次大小，分批次，每次提交的 id 数
    :param max_workers: 并发工作数，如果为 None 或者 <= 0，则自动确定
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数
    """
    return update_abstract(
        client, 
        ids, # type: ignore
        method="fs_category_shortcut_set", 
        value=set, 
        batch_size=batch_size, 
        max_workers=max_workers, 
        async_=async_, # type: ignore
        **request_kwargs, 
    )


@overload
def batch_unstar(
    client: str | PathLike | P115Client, 
    /, 
    batch_size: int = 10_000, 
    ensure_file: None | bool = None, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
):
    ...
@overload
def batch_unstar(
    client: str | PathLike | P115Client, 
    /, 
    batch_size: int = 10_000, 
    ensure_file: None | bool = None, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine:
    ...
def batch_unstar(
    client: str | PathLike | P115Client, 
    /, 
    batch_size: int = 10_000, 
    ensure_file: None | bool = None, 
    max_workers: None | int = None, 
    app: str = "web", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
):
    """批量对一批文件或目录取消星标

    :param client: 115 客户端或 cookies
    :param batch_size: 批次大小，分批次，每次提交的 id 数
    :param ensure_file: 是否确保为文件

        - True: 必须是文件
        - False: 必须是目录
        - None: 可以是目录或文件

    :param max_workers: 并发工作数，如果为 None 或者 <= 0，则自动确定
    :param app: 使用此设备的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    def get_id(info: dict, /) -> int:
        for k in ("file_id", "category_id", "fid", "cid"):
            if k in info:
                return int(info[k])
        raise KeyError
    def gen_step():
        from .iterdir import _iter_fs_files
        yield update_star(
            client, 
            do_map(get_id, _iter_fs_files(
                client, 
                payload={
                    "cid": 0, "count_folders": 1, "cur": 0, "fc_mix": 0, 
                    "offset": 0, "show_dir": 1, "star": 1
                }, 
                ensure_file=ensure_file, 
                app=app, 
                cooldown=0.5, 
                async_=async_, 
                **request_kwargs, 
            )), 
            star=False, 
            batch_size=batch_size, 
            max_workers=max_workers, 
            app=app, 
            async_=async_, 
            **request_kwargs, 
        )
    return run_gen_step(gen_step, async_)


@overload
def update_name(
    client: str | PathLike | P115Client, 
    id_name_pairs: Iterable[tuple[int | str, str]], 
    /, 
    batch_size: int = 10_000, 
    post_event_type: None | Literal["doc", "img"] = "img", 
    app: str = "web", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> list[Future]:
    ...
@overload
def update_name(
    client: str | PathLike | P115Client, 
    id_name_pairs: Iterable[tuple[int | str, str]], 
    /, 
    batch_size: int = 10_000, 
    post_event_type: None | Literal["doc", "img"] = "img", 
    app: str = "web", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, list[AsyncFuture]]:
    ...
def update_name(
    client: str | PathLike | P115Client, 
    id_name_pairs: Iterable[tuple[int | str, str]], 
    /, 
    batch_size: int = 10_000, 
    post_event_type: None | Literal["doc", "img"] = "img", 
    app: str = "web", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> list[Future] | Coroutine[Any, Any, list[AsyncFuture]]:
    """批量给文件或目录设置名字

    :param client: 115 客户端或 cookies
    :param id_name_pairs: 一堆文件或目录的 id 到新名字的元组
    :param batch_size: 批次大小，分批次，每次提交的任务数
    :param post_event_type: 推送事件类型，如果为 None，则不推送
    :param app: 使用此设备的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 返回推送事件的 Future 对象列表（因为是并发执行的）
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if app in ("", "web", "desktop", "harmony"):
        method: Callable = client.fs_rename
    else:
        method = client.fs_rename_app
        request_kwargs["app"] = app
    if async_:
        run: Callable = run_as_async
    else:
        run = run_as_thread
    def gen_step():
        futures: list = []
        for batch in batched(id_name_pairs, batch_size):
            resp = yield method(
                batch, 
                async_=async_, 
                **request_kwargs, 
            )
            check_response(resp)
            if post_event_type and resp["data"]:
                futures.append(run(
                    post_event, 
                    client, 
                    resp["data"].keys(), 
                    type=post_event_type, 
                    batch_size=batch_size, 
                    max_workers=1, 
                    app=app, 
                    async_=async_, 
                    **request_kwargs, 
                ))
        return futures
    return run_gen_step(gen_step, async_)


# TODO: 是否能批量推送 "browse_audio" 或 "browse_video" 事件？
@overload
def post_event(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str], 
    /, 
    type: Literal["doc", "img"] = "doc", 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "android", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> None:
    ...
@overload
def post_event(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    type: Literal["doc", "img"] = "doc", 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "android", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine:
    ...
def post_event(
    client: str | PathLike | P115Client, 
    ids: Iterable[int | str] | AsyncIterable[int | str], 
    /, 
    type: Literal["doc", "img"] = "doc", 
    batch_size: int = 10_000, 
    max_workers: None | int = None, 
    app: str = "android", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> None | Coroutine:
    """批量将文件或目录推送事件

    :param client: 115 客户端或 cookies
    :param ids: 一组文件或目录的 id 或 pickcode
    :param type: 事件类型

        - "doc": 推送 "browse_document" 事件
        - "img": 推送 "browse_image" 事件

    :param batch_size: 批次大小，分批次，每次提交的 id 数
    :param max_workers: 并发工作数，如果为 None 或者 <= 0，则自动确定
    :param app: 使用此设备的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if type == "doc":
        post = client.life_behavior_doc_post_app
    else:
        post = client.life_behavior_img_post_app
    def call(batch, /):
        return check_response(post(
            batch, 
            app=app, 
            async_=async_, 
            request_kwargs=request_kwargs, 
        ))
    def gen_step():
        yield through(conmap(
            call, 
            chunked(do_map(to_id, ids), batch_size), 
            max_workers=max_workers, 
            async_=async_, 
        ))
    return run_gen_step(gen_step, async_)


@overload
def batch_makedir(
    client: str | PathLike | P115Client, 
    pairs: Iterable[str | tuple[int | str, str]], 
    /, 
    pid: int | str = 0, 
    contain_dir: bool = False, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator[tuple[str | tuple[int | str, str], dict]]:
    ...
@overload
def batch_makedir(
    client: str | PathLike | P115Client, 
    pairs: Iterable[str | tuple[int | str, str]], 
    /, 
    pid: int | str = 0, 
    contain_dir: bool = False, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator[tuple[str | tuple[int | str, str], dict]]:
    ...
def batch_makedir(
    client: str | PathLike | P115Client, 
    pairs: Iterable[str | tuple[int | str, str]], 
    /, 
    pid: int | str = 0, 
    contain_dir: bool = False, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> Iterator[tuple[str | tuple[int | str, str], dict]] | AsyncIterator[tuple[str | tuple[int | str, str], dict]]:
    """批量创建目录

    :param client: 115 客户端或 cookies
    :param pairs: 一系列的 **名字或相对路径** 或者 (**目录的 id 或 pickcode**, **名字或相对路径**) 的 2 元组
    :param pid: 目录的 id 或 pickcode，如果输入的是 **名字或相对路径**，则创建在此目录下
    :param contain_dir: 如果为 True，则要创建的是相对路径，否则就是一个文件（即使其中包含 "/"）
    :param max_workers: 并发工作数，如果为 None 或者 <= 0，则自动确定
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 迭代器，产生 (**每项输入**, **相应的接口响应**) 的 2 元组
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    pid = to_id(pid)
    if contain_dir:
        makedir = client.fs_makedirs_app
    else:
        makedir = client.fs_mkdir_app
    @as_gen_step
    def call[T: (str, tuple[int | str, str])](pair: T, /):
        if isinstance(pair, tuple):
            cid, name = pair
            cid = to_id(cid)
        else:
            cid = pid
            name = pair
        return pair, (yield makedir(name, pid=cid, async_=async_, **request_kwargs))
    return conmap(
        call, 
        pairs, 
        max_workers=max_workers, 
        async_=async_, 
    )


@overload
def copyfile(
    client: str | PathLike | P115Client | P115OpenClient, 
    id: int | str | Mapping, 
    /, 
    name: str = "", 
    pid: None | int | str = None, 
    app: str = "web", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def copyfile(
    client: str | PathLike | P115Client | P115OpenClient, 
    id: int | str | Mapping, 
    /, 
    name: str = "", 
    pid: None | int | str = None, 
    app: str = "web", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def copyfile(
    client: str | PathLike | P115Client | P115OpenClient, 
    id: int | str | Mapping, 
    /, 
    name: str = "", 
    pid: None | int | str = None, 
    app: str = "web", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """复制文件到目标目录下

    .. note::
        如果复制后的文件名相同，则使用复制接口，否则使用上传接口

    :param client: 115 客户端或 cookies
    :param id: 文件的 id、pickcode 或信息字典
    :param name: 复制后的新名字，如果为空则名字相同
    :param pid: 目录的 id 或 pickcode，如果为 None，则在同一目录下
    :param app: 使用指定 app（设备）的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口响应信息
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if not isinstance(client, P115Client) or app == "open":
        get_url: Callable = client.download_url_open
        fs_copy: Callable = client.fs_copy_open
        upload_file_init: Callable = client.upload_file_init_open
    else:
        get_url = client.download_url
        upload_file_init = client.upload_file_init
        if app in ("", "web", "desktop", "harmony"):
            fs_copy = client.fs_copy
        else:
            if isinstance(id, Mapping):
                get_url = partial(get_url, app=app)
            fs_copy = partial(client.fs_copy_app, app=app)
    def gen_step():
        nonlocal pid, name
        url: None | P115URL = None
        if isinstance(id, Mapping):
            attr: Mapping = id
        else:
            url = yield get_url(
                client.to_pickcode(id), 
                async_=async_, 
                **request_kwargs, 
            )
            attr = url.__dict__
        if pid is None:
            pid = attr["parent_id"]
        else:
            pid = client.to_id(pid)
        if not name:
            name = attr["name"]
        if attr["name"] == name:
            if pid == attr["parent_id"]:
                return {"state": True}
            return check_response(fs_copy(attr["id"], pid=pid, async_=async_, **request_kwargs))
        else:
            @as_gen_step
            def read_range_bytes_or_hash(sign_check: str, /):
                nonlocal url
                if url is None:
                    url = yield get_url(
                        attr.get("pickcode") or client.to_pickcode(attr["id"]), 
                        async_=async_, 
                        **request_kwargs, 
                    )
                return client.request(
                    url, 
                    async_=async_, 
                    **{
                        **request_kwargs, 
                        "headers": {**url.headers, "range": "bytes="+sign_check}, 
                        "parse": False, 
                    }
                )
            return check_response(upload_file_init(
                filename=name, 
                filesha1=attr["sha1"], 
                filesize=attr["size"], 
                read_range_bytes_or_hash=read_range_bytes_or_hash, 
                pid=pid, 
                async_=async_, 
                **request_kwargs, 
            ))
    return run_gen_step(gen_step, async_)


@overload
def renamefile(
    client: str | PathLike | P115Client | P115OpenClient, 
    id: int | str | Mapping, 
    /, 
    name: str = "", 
    pid: None | int | str = None, 
    app: str = "web", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def renamefile(
    client: str | PathLike | P115Client | P115OpenClient, 
    id: int | str | Mapping, 
    /, 
    name: str = "", 
    pid: None | int | str = None, 
    app: str = "web", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def renamefile(
    client: str | PathLike | P115Client | P115OpenClient, 
    id: int | str | Mapping, 
    /, 
    name: str = "", 
    pid: None | int | str = None, 
    app: str = "web", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """移动文件到目标目录下

    .. note::
        如果移动后的文件名相同或扩展名相同，则使用移动接口，否则使用上传接口

    :param client: 115 客户端或 cookies
    :param id: 文件的 id、pickcode 或信息字典
    :param name: 移动后的新名字，如果为空则名字相同
    :param pid: 目录的 id 或 pickcode，如果为 None，则不进行移动
    :param app: 使用指定 app（设备）的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口响应信息
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if not isinstance(client, P115Client) or app == "open":
        get_url: Callable = client.download_url_open
        fs_move: Callable = client.fs_move_open
        upload_file_init: Callable = client.upload_file_init_open
        fs_delete: Callable = client.fs_delete_open
        fs_rename: Callable = client.fs_rename_open
    else:
        get_url = client.download_url
        upload_file_init = client.upload_file_init
        if app in ("", "web", "desktop", "harmony"):
            fs_move = client.fs_move
            fs_delete = client.fs_delete
            fs_rename = client.fs_rename
        else:
            if isinstance(id, Mapping):
                get_url = partial(get_url, app=app)
            fs_move = partial(client.fs_move_app, app=app)
            fs_delete = partial(client.fs_delete_app, app=app)
            fs_rename = partial(client.fs_rename_app, app=app)
    def gen_step():
        nonlocal pid, name
        url: None | P115URL = None
        if isinstance(id, Mapping):
            attr: Mapping = id
        else:
            url = yield get_url(
                client.to_pickcode(id), 
                async_=async_, 
                **request_kwargs, 
            )
            attr = url.__dict__
        if pid is None:
            pid = attr["parent_id"]
        else:
            pid = client.to_id(pid)
        if not name:
            name = attr["name"]
        is_same_name = attr["name"] == name
        is_same_ext = attr["name"].rpartition(".")[-1] == name.rpartition(".")[-1]
        if is_same_name or is_same_ext:
            if pid == attr["parent_id"]:
                if is_same_name:
                    return {"state": True}
            else:
                resp = yield check_response(fs_move(attr["id"], pid=pid, async_=async_, **request_kwargs))
                check_response(resp)
                if is_same_name:
                    return resp
            return check_response(fs_rename((attr["id"], name), async_=async_, **request_kwargs))
        else:
            @as_gen_step
            def read_range_bytes_or_hash(sign_check: str, /):
                nonlocal url
                if url is None:
                    url = yield get_url(
                        attr.get("pickcode") or client.to_pickcode(attr["id"]), 
                        async_=async_, 
                        **request_kwargs, 
                    )
                return client.request(
                    url, 
                    async_=async_, 
                    **{
                        **request_kwargs, 
                        "headers": {**url.headers, "range": "bytes="+sign_check}, 
                        "parse": False, 
                    }
                )
            resp = yield upload_file_init(
                filename=name, 
                filesha1=attr["sha1"], 
                filesize=attr["size"], 
                read_range_bytes_or_hash=read_range_bytes_or_hash, 
                pid=pid, 
                async_=async_, 
                **request_kwargs, 
            )
            check_response(resp)
            return check_response(fs_delete(attr["id"], async_=async_, **request_kwargs))
    return run_gen_step(gen_step, async_)


@overload
def transferfile(
    client_from: str | PathLike | P115Client | P115OpenClient, 
    client_to: str | PathLike | P115Client | P115OpenClient, 
    id: int | str | Mapping, 
    /, 
    name: str = "", 
    pid: int | str = 0, 
    app: str = "web", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def transferfile(
    client_from: str | PathLike | P115Client | P115OpenClient, 
    client_to: str | PathLike | P115Client | P115OpenClient, 
    id: int | str | Mapping, 
    /, 
    name: str = "", 
    pid: int | str = 0, 
    app: str = "web", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def transferfile(
    client_from: str | PathLike | P115Client | P115OpenClient, 
    client_to: str | PathLike | P115Client | P115OpenClient, 
    id: int | str | Mapping, 
    /, 
    name: str = "", 
    pid: int | str = 0, 
    app: str = "web", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """从一个 115 网盘，转移文件到另一个 115 网盘

    :param client_from: 115 客户端或 cookies，文件来自这个网盘
    :param client_to: 115 客户端或 cookies，文件去往这个网盘
    :param id: 文件的 id、pickcode 或信息字典，关联在 `client_from`
    :param name: 复制后的新名字，如果为空则名字相同，关联在 `client_to`
    :param pid: 目录的 id 或 pickcode，，关联在 `client_to`
    :param app: 使用指定 app（设备）的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口响应信息
    """
    if isinstance(client_from, (str, PathLike)):
        client_from = P115Client(client_from, check_for_relogin=True)
    if isinstance(client_to, (str, PathLike)):
        client_to = P115Client(client_to, check_for_relogin=True)
    if not isinstance(client_from, P115Client) or app == "open":
        get_url: Callable = client_from.download_url_open
    else:
        get_url = client_from.download_url
        if isinstance(id, Mapping):
            get_url = partial(get_url, app=app)
    if not isinstance(client_to, P115Client) or app == "open":
        upload_file_init: Callable = client_to.upload_file_init_open
    else:
        upload_file_init = client_to.upload_file_init
    def gen_step():
        nonlocal pid, name
        url: None | P115URL = None
        if isinstance(id, Mapping):
            attr: Mapping = id
        else:
            url = yield get_url(
                client_from.to_pickcode(id), 
                async_=async_, 
                **request_kwargs, 
            )
            attr = url.__dict__
        pid = client_to.to_id(pid)
        if not name:
            name = attr["name"]
        @as_gen_step
        def read_range_bytes_or_hash(sign_check: str, /):
            nonlocal url
            if url is None:
                url = yield get_url(
                    attr.get("pickcode") or client_from.to_pickcode(attr["id"]), 
                    async_=async_, 
                    **request_kwargs, 
                )
            return client_to.request(
                url, 
                async_=async_, 
                **{
                    **request_kwargs, 
                    "headers": {**url.headers, "range": "bytes="+sign_check}, 
                    "parse": False, 
                }
            )
        return check_response(upload_file_init(
            filename=name, 
            filesha1=attr["sha1"], 
            filesize=attr["size"], 
            read_range_bytes_or_hash=read_range_bytes_or_hash, 
            pid=pid, 
            async_=async_, 
            **request_kwargs, 
        ))
    return run_gen_step(gen_step, async_)

# TODO: 上面这些，有些要支持 open 接口



#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = [
    "HISTORY_NAME_TO_TYPE", "HISTORY_TYPE_TO_NAME", 
    "iter_history_once", "iter_history", "iter_history_list", 
]
__doc__ = "这个模块提供了一些和 115 的历史记录有关的函数"

from asyncio import sleep as async_sleep
from collections.abc import AsyncIterator, Container, Iterator
from functools import partial
from itertools import cycle
from os import PathLike
from time import time, sleep
from typing import overload, Final, Literal

from iterutils import run_gen_step_iter, with_iter_next, Yield

from ..client import check_response, P115Client


#: 115 生活操作事件名称到类型的映射
HISTORY_NAME_TO_TYPE: Final = {
    "all": 0, 
    "offline_download": 2, 
    "browse_video": 3, 
    "upload": 4, 
    "receive": 7, 
    "move": 8, 
}
#: 115 生活操作事件类型到名称的映射
HISTORY_TYPE_TO_NAME: Final = {v: k for k, v in HISTORY_NAME_TO_TYPE.items()}


@overload
def iter_history_once(
    client: str | PathLike | P115Client, 
    from_id: int = 0, 
    from_time: float = 0, 
    type: int | str = 0, 
    ignore_types: None | Container[int] = None, 
    first_batch_size = 0, 
    app: str = "web", 
    cooldown: int | float = 0, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator[dict]:
    ...
@overload
def iter_history_once(
    client: str | PathLike | P115Client, 
    from_id: int = 0, 
    from_time: float = 0, 
    type: int | str = 0, 
    ignore_types: None | Container[int] = None, 
    first_batch_size = 0, 
    app: str = "web", 
    cooldown: int | float = 0, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator[dict]:
    ...
def iter_history_once(
    client: str | PathLike | P115Client, 
    from_id: int = 0, 
    from_time: float = 0, 
    type: int | str = 0, 
    ignore_types: None | Container[int] = None, 
    first_batch_size = 0, 
    app: str = "web", 
    cooldown: int | float = 0, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> AsyncIterator[dict] | Iterator[dict]:
    """拉取一组 115 的历史记录

    .. note::
        当你指定有 ``from_id != 0`` 时，如果 from_time 为 0，则自动重设为 -1

    :param client: 115 客户端或 cookies
    :param from_id: 开始的事件 id （不含）
    :param from_time: 开始时间（含），若为 0 则从当前时间开始，若 < 0 则从最早开始
    :param type: 拉取指定类型的历史记录（？？表示还未搞清楚），多个用逗号 "," 隔开

        - 全部: 0
        - ？？: 1
        - 离线下载: 2
        - 播放视频: 3
        - 上传: 4
        - ？？: 5
        - ？？: 6
        - 接收: 7
        - 移动: 8

    :param ignore_types: 一组要被忽略的操作事件类型代码，仅当 `type` 为空时生效
    :param first_batch_size: 首批的拉取数目
    :param app: 使用某个 app （设备）的接口
    :param cooldown: 冷却时间，大于 0 时，两次接口调用之间至少间隔这么多秒
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 迭代器，产生 115 的历史记录数据字典
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if app in ("", "web", "desktop", "harmony"):
        history_list = partial(client.fs_history_list, **request_kwargs)
    else:
        request_kwargs.setdefault("base_url", cycle(("http://pro.api.115.com", "https://proapi.115.com")).__next__)
        history_list = partial(client.fs_history_list_app, app=app, **request_kwargs)
    if first_batch_size <= 0:
        first_batch_size = 64 if from_time or from_id else 1150
    if from_id and not from_time:
        from_time = -1
    def gen_step():
        payload = {"type": type, "limit": first_batch_size, "offset": 0}
        seen: set[int] = set()
        seen_add = seen.add
        ts_last_call = time()
        resp = yield history_list(payload, async_=async_)
        events = check_response(resp)["data"]["list"]
        payload["limit"] = 1150
        offset = 0
        while events:
            for event in events:
                event_id = int(event["id"])
                if (from_id and event_id <= from_id or 
                    from_time and "update_time" in event and int(event["update_time"]) < from_time
                ):
                    return
                event_type = event["type"]
                if event_id not in seen:
                    if type or not ignore_types or event_type not in ignore_types:
                        event["event_name"] = HISTORY_TYPE_TO_NAME.get(event_type, "")
                        yield Yield(event)
                    seen_add(event_id)
            offset += len(events)
            if offset >= int(resp["data"]["total"]):
                return
            payload["offset"] = offset
            if cooldown > 0 and (delta := ts_last_call + cooldown - time()) > 0:
                if async_:
                    yield async_sleep(delta)
                else:
                    sleep(delta)
            ts_last_call = time()
            resp = yield history_list(payload, async_=async_)
            events = check_response(resp)["data"]["list"]
    return run_gen_step_iter(gen_step, async_)


@overload
def iter_history(
    client: str | PathLike | P115Client, 
    from_id: int = 0, 
    from_time: float = 0, 
    type: int | str = 0, 
    ignore_types: None | Container[int] = None, 
    app: str = "web", 
    cooldown: int | float = 0, 
    interval: int | float = 0, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator[dict]:
    ...
@overload
def iter_history(
    client: str | PathLike | P115Client, 
    from_id: int = 0, 
    from_time: float = 0, 
    type: int | str = 0, 
    ignore_types: None | Container[int] = None, 
    app: str = "web", 
    cooldown: int | float = 0, 
    interval: int | float = 0, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator[dict]:
    ...
def iter_history(
    client: str | PathLike | P115Client, 
    from_id: int = 0, 
    from_time: float = 0, 
    type: int | str = 0, 
    ignore_types: None | Container[int] = None, 
    app: str = "web", 
    cooldown: int | float = 0, 
    interval: int | float = 0, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> AsyncIterator[dict] | Iterator[dict]:
    """持续拉取 115 的历史记录

    .. note::
        当你指定有 ``from_id != 0`` 时，如果 from_time 为 0，则自动重设为 -1

    :param client: 115 客户端或 cookies
    :param from_id: 开始的事件 id （不含）
    :param from_time: 开始时间（含），若为 0 则从当前时间开始，若 < 0 则从最早开始
    :param type: 拉取指定类型的历史记录（？？表示还未搞清楚），多个用逗号 "," 隔开

        - 全部: 0
        - ？？: 1
        - 离线下载: 2
        - 播放视频: 3
        - 上传: 4
        - ？？: 5
        - ？？: 6
        - 接收: 7
        - 移动: 8

    :param ignore_types: 一组要被忽略的操作事件类型代码，仅当 `type` 为空时生效
    :param cooldown: 冷却时间，大于 0 时，两次接口调用之间至少间隔这么多秒
    :param interval: 两个批量拉取之间的睡眠时间间隔，如果小于等于 0，则不睡眠
    :param app: 使用某个 app （设备）的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 迭代器，产生 115 的历史记录数据字典
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if from_id and not from_time:
        from_time = -1
    def gen_step():
        nonlocal from_time, from_id
        if from_time == 0:
            from_time = time()
        first_loop = True
        while True:
            if first_loop:
                first_loop = False
            elif interval > 0:
                if async_:
                    yield async_sleep(interval)
                else:
                    sleep(interval)
            with with_iter_next(iter_history_once(
                client, 
                from_id=from_id, 
                from_time=from_time, 
                type=type, 
                app=app, 
                cooldown=cooldown, 
                async_=async_, 
                **request_kwargs, 
            )) as get_next:
                sub_first_loop = True
                while True:
                    event = yield get_next()
                    if sub_first_loop:
                        from_id = int(event["id"])
                        if "update_time" in event:
                            from_time = int(event["update_time"])
                        else:
                            from_time = 0
                        sub_first_loop = False
                    if not type and ignore_types and event["type"] in ignore_types:
                        continue
                    yield Yield(event)
    return run_gen_step_iter(gen_step, async_)


@overload
def iter_history_list(
    client: str | PathLike | P115Client, 
    from_id: int = 0, 
    from_time: float = 0, 
    type: int | str = 0, 
    ignore_types: None | Container[int] = None, 
    app: str = "web", 
    cooldown: int | float = 0, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator[dict]:
    ...
@overload
def iter_history_list(
    client: str | PathLike | P115Client, 
    from_id: int = 0, 
    from_time: float = 0, 
    type: int | str = 0, 
    ignore_types: None | Container[int] = None, 
    app: str = "web", 
    cooldown: int | float = 0, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator[dict]:
    ...
def iter_history_list(
    client: str | PathLike | P115Client, 
    from_id: int = 0, 
    from_time: float = 0, 
    type: int | str = 0, 
    ignore_types: None | Container[int] = None, 
    app: str = "web", 
    cooldown: int | float = 0, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> AsyncIterator[dict] | Iterator[dict]:
    """持续拉取 115 的历史记录

    .. note::
        当你指定有 ``from_id != 0`` 时，如果 from_time 为 0，则自动重设为 -1

    :param client: 115 客户端或 cookies
    :param from_id: 开始的事件 id （不含）
    :param from_time: 开始时间（含），若为 0 则从当前时间开始，若 < 0 则从最早开始
    :param type: 拉取指定类型的历史记录（？？表示还未搞清楚），多个用逗号 "," 隔开

        - 全部: 0
        - ？？: 1
        - 离线下载: 2
        - 播放视频: 3
        - 上传: 4
        - ？？: 5
        - ？？: 6
        - 接收: 7
        - 移动: 8

    :param ignore_types: 一组要被忽略的操作事件类型代码，仅当 `type` 为空时生效
    :param cooldown: 冷却时间，大于 0 时，两次接口调用之间至少间隔这么多秒
    :param app: 使用某个 app （设备）的接口
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 迭代器，产生 115 的历史记录数据字典
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if from_id and not from_time:
        from_time = -1
    def gen_step():
        nonlocal from_time, from_id
        if from_time == 0:
            from_time = time()
        while True:
            ls: list[dict] = []
            push = ls.append
            with with_iter_next(iter_history_once(
                client, 
                from_id=from_id, 
                from_time=from_time, 
                type=type, 
                app=app, 
                cooldown=cooldown, 
                async_=async_, 
                **request_kwargs, 
            )) as get_next:
                first_loop = True
                while True:
                    event = yield get_next()
                    if first_loop:
                        from_id = int(event["id"])
                        if "update_time" in event:
                            from_time = int(event["update_time"])
                        else:
                            from_time = 0
                        first_loop = False
                    if not type and ignore_types and event["type"] in ignore_types:
                        continue
                    push(event)
            yield Yield(ls)
    return run_gen_step_iter(gen_step, async_)




#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__version__ = (0, 0, 8)

from .client import *
from .const import *
from .exception import *
from .fs import *
from .type import *
from . import util



#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = [
    "DirNode", "P115UID", "P115Cookies", "P115DictAttrLikeMixin", 
    "P115DictAttrLike", "P115ID", "P115StrID", "P115URL", 
]

from datetime import datetime, timedelta
from functools import cached_property
from http.cookiejar import CookieJar, Cookie
from http.cookies import BaseCookie, Morsel
from re import compile as re_compile
from time import time
from types import MappingProxyType
from typing import Any, Final, NamedTuple, Self

from cookietools import cookies_to_dict, cookies_to_str
from integer_tool import try_parse_int
from undefined import undefined

from .const import SSOENT_TO_APP


CRE_UID_FORMAT_match: Final = re_compile("(?P<user_id>[1-9][0-9]*)_(?P<login_ssoent>[A-Z][1-9][0-9]*)_(?P<login_timestamp>[1-9][0-9]{9,})").fullmatch
CRE_CID_FORMAT_match: Final = re_compile("[0-9a-f]{32}").fullmatch
CRE_SEID_FORMAT_match: Final = re_compile("[0-9a-f]{120}").fullmatch


class DirNode(NamedTuple):
    """用来保存某个 id 对应的 name 和 parent_id 的元组
    """
    name: str
    parent_id: int


class P115UID(str):

    def __init__(self, /, *a, **k):
        if m := CRE_UID_FORMAT_match(self):
            self.__dict__.update((k, try_parse_int(v)) for k, v in m.groupdict().items())

    def __repr__(self, /) -> str:
        cls = type(self)
        return f"{cls.__module__}.{cls.__qualname__}({super().__repr__()})"

    @cached_property
    def user_id(self, /) -> int:
        return 0

    @cached_property
    def login_ssoent(self, /) -> str:
        return ""

    @cached_property
    def login_timestamp(self, /) -> int:
        return 0


class P115Cookies(str):
    """cookies 的封装
    """
    __last_new_instance__ = None

    def __new__(cls, cookies, /):
        def predicate(_, val, /):
            domain: None | str = None
            if isinstance(val, Cookie):
                domain = val.domain
            elif isinstance(val, Morsel):
                domain = val["domain"]
            return not domain or domain == "115.com" or domain.endswith(".115.com")
        cookies = cookies_to_str(cookies, predicate)
        if cookies == cls.__last_new_instance__:
            return cls.__last_new_instance__
        else:
            inst = cls.__last_new_instance__ = super().__new__(cls, cookies)
            return inst

    def __deepcopy__(self, /, memo) -> Self:
        return self

    def __getattr__(self, attr: str, /):
        try:
            return self.mapping[attr]
        except KeyError as e:
            raise AttributeError(attr) from e

    def __getitem__(self, key, /): # type: ignore
        if isinstance(key, str):
            return self.mapping[key]
        return super().__getitem__(key)

    def __repr__(self, /) -> str:
        cls = type(self)
        return f"{cls.__module__}.{cls.__qualname__}({super().__repr__()})"

    def __setattr__(self, attr, value, /):
        raise TypeError("can't set attribute")

    @cached_property
    def mapping(self, /) -> MappingProxyType:
        return MappingProxyType(cookies_to_dict(self))

    @cached_property
    def uid(self, /) -> P115UID:
        "UID 字段"
        return P115UID(self.mapping.get("UID") or "")

    @cached_property
    def cid(self, /) -> str:
        "CID 字段"
        return self.mapping.get("CID") or ""

    @cached_property
    def kid(self, /) -> str:
        "KID 字段"
        return self.mapping.get("KID") or ""

    @cached_property
    def seid(self, /) -> str:
        "SEID 字段"
        return self.mapping.get("SEID") or ""

    @property
    def user_id(self, /) -> int:
        "用户 ID"
        return self.uid.user_id

    @property
    def login_ssoent(self, /) -> str:
        "登录设备标识"
        return self.uid.login_ssoent

    ssoent = login_ssoent

    @cached_property
    def login_app(self, /) -> None | str:
        "登录设备名"
        return SSOENT_TO_APP.get(self.login_ssoent)

    @property
    def app(self, /) -> None | str:
        "登录设备名"
        return self.login_app

    @property
    def login_timestamp(self, /) -> int:
        "登录时间戳"
        return self.uid.login_timestamp

    timestamp = login_timestamp

    @cached_property
    def datetime(self, /) -> datetime:
        "登录时间"
        return datetime.fromtimestamp(self.login_timestamp)

    @property
    def time_elapsed(self, /) -> float:
        "从登录到现在过了多少时间，单位：秒"
        return time() - self.login_timestamp

    @property
    def time_delta(self, /) -> timedelta:
        "从登录到现在过了多少时间"
        return datetime.now() - self.datetime

    @cached_property
    def is_well_formed(self, /) -> bool:
        "是否格式良好，即需要的字段都具备而且都符合格式"
        return (
            CRE_UID_FORMAT_match(self.uid) and 
            CRE_CID_FORMAT_match(self.cid) and 
            CRE_SEID_FORMAT_match(self.seid)
        ) is not None

    @cached_property
    def cookies(self, /) -> str:
        """115 登录的 cookies，包含 UID、CID、KID 和 SEID 这 4 个字段
        """
        cookies = f"UID={self.uid}; CID={self.cid}; SEID={self.seid}"
        if "KID" in self.mapping:
            cookies += f"; KID={self.mapping['KID']}"
        return cookies

    @classmethod
    def from_dict(cls, cookies: dict[str, str], /) -> Self:
        return cls("; ".join(f"{key}={val}" for key, val in cookies.items()))

    @classmethod
    def from_cookiejar(cls, cookiejar: CookieJar, /) -> Self:
        return cls("; ".join(
            f"{cookie.name}={cookie.value}" 
            for cookie in cookiejar 
            if not (domain := cookie.domain) or domain == "115.com" or domain.endswith(".115.com")
        ))

    @classmethod
    def from_simple_cookie(cls, cookies: BaseCookie, /) -> Self:
        return cls("; ".join(
            f"{name}={cookie.value}" 
            for name, cookie in cookies.items() 
            if not (domain := cookie["domain"]) or domain == "115.com" or domain.endswith(".115.com")
        ))


class P115DictAttrLikeMixin:

    def __getattr__(self, attr: str, /):
        try:
            return self.__dict__[attr]
        except KeyError as e:
            raise AttributeError(attr) from e

    def __delitem__(self, key: str, /):
        del self.__dict__[key]

    def __getitem__(self, key, /):
        try:
            if isinstance(key, str):
                return self.__dict__[key]
        except KeyError:
            return super().__getitem__(key) # type: ignore

    def __setitem__(self, key: str, val, /):
        self.__dict__[key] = val

    def __repr__(self, /) -> str:
        cls = type(self)
        if (module := cls.__module__) not in ("__main__", "builtins"):
            name = cls.__qualname__
        else:
            name = f"{module}.{cls.__qualname__}"
        return f"{name}({super().__repr__()}, {self.__dict__!r})"

    @property
    def mapping(self, /) -> dict[str, Any]:
        return self.__dict__

    def get(self, key, /, default=None):
        return self.__dict__.get(key, default)

    def items(self, /):
        return self.__dict__.items()

    def keys(self, /):
        return self.__dict__.keys()

    def values(self, /):
        return self.__dict__.values()


class P115DictAttrLike(P115DictAttrLikeMixin):

    def __new__(cls, val: Any = undefined, /, *args, **kwds):
        if val is undefined:
            return super().__new__(cls)
        else:
            return super().__new__(cls, val) # type: ignore

    def __init__(self, val: Any = undefined, /, *args, **kwds):
        self.__dict__.update(*args, **kwds)

    @classmethod
    def of(cls, val: Any = undefined, /, ns: None | dict = None) -> Self:
        if val is undefined:
            self = cls.__new__(cls)
        else:
            self = cls.__new__(cls, val)
        if ns is not None:
            self.__dict__ = ns
        return self

    @classmethod
    def derive(cls, base: type, name: str = "", /, **ns) -> type[Self]:
        return type(name, (cls, base), ns)

    @classmethod
    def derive_backend(cls, base: type, name: str = "", /, **ns) -> type[Self]:
        return type(name, (base, cls), ns)


class P115ID(P115DictAttrLike, int):
    """整数 id 的封装
    """
    def __str__(self, /) -> str:
        return int.__repr__(self)


class P115StrID(P115DictAttrLike, str):
    """字符串 id 的封装
    """


class P115URL(P115DictAttrLike, str):
    """下载链接的封装
    """
    def geturl(self, /) -> str:
        return str(self)

    url = property(geturl)




#!/usr/bin/env python3
# encoding: utf-8

from .fs_base import *
from .fs import *
from .fs_share import *
from .fs_zip import *

def __getattr__(attr, /):
    if attr == "__all__":
        from . import fs_base, fs, fs_share, fs_zip
        return fs_base.__all__ + fs.__all__ + fs_share.__all__ + fs_zip.__all__
    raise AttributeError(attr)



#!/usr/bin/env python3
# encoding: utf-8

from __future__ import annotations

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = [
    "iter_115_to_115", "iter_115_to_115_resume", "upload_init", 
    "P115MultipartUpload", 
]
__doc__ = "这个模块提供了一些和上传有关的函数"

from collections.abc import (
    AsyncIterable, AsyncIterator, Awaitable, Buffer, Iterable, Callable, 
    Coroutine, Iterator, 
)
from inspect import isawaitable
from itertools import count, dropwhile
from os import fsdecode, PathLike
from typing import cast, overload, Any, Literal

from asynctools import to_list
from concurrenttools import threadpool_map, taskgroup_map, Return
from dicttools import dict_map
from filewrap import SupportsRead
from http_request import SupportsGeturl
from iterutils import (
    as_gen_step, foreach, run_gen_step, run_gen_step_iter, 
    with_iter_next, YieldFrom, 
)
from p115oss import (
    oss_upload_init, oss_multipart_upload_init, oss_multipart_upload_complete, 
    oss_multipart_upload_url, oss_multipart_part_iter, oss_multipart_upload_part_iter, 
)
from p115pickcode import to_id
from yarl import URL

from ..client import check_response, P115Client, P115OpenClient
from .attr import normalize_attr_simple
from .download import iter_download_files
from .iterdir import iterdir, iter_files_with_path, unescape_115_charref


# TODO: 需要优化，减少代码量
# TODO: 支持 open 接口
# TODO: 再支持一个方法，目标 115 并不提供 client，只有 user_id 和 user_key
@overload
def iter_115_to_115(
    from_client: P115Client, 
    to_client: P115Client, 
    from_cid: int | str = 0, 
    to_pid: int | str = 0, 
    max_workers: int = 8, 
    with_root: bool = True, 
    use_iter_files: bool = False, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator[dict]:
    ...
@overload
def iter_115_to_115(
    from_client: P115Client, 
    to_client: P115Client, 
    from_cid: int | str = 0, 
    to_pid: int | str = 0, 
    max_workers: int = 8, 
    with_root: bool = True, 
    use_iter_files: bool = False, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator[dict]:
    ...
def iter_115_to_115(
    from_client: P115Client, 
    to_client: P115Client, 
    from_cid: int | str = 0, 
    to_pid: int | str = 0, 
    max_workers: int = 8, 
    with_root: bool = True, 
    use_iter_files: bool = False, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> Iterator[dict] | AsyncIterator[dict]:
    """从 115 传到 115

    :param from_client: 来源 115 客户端对象
    :param to_client: 去向 115 客户端对象
    :param from_cid: 来源 115 的目录 id 或 pickcode
    :param to_pid: 去向 115 的父目录 id 或 pickcode
    :param max_workers: 最大并发数
    :param with_root: 是否保留 `from_cid` 对应的目录名（如果为 False，则会少 1 级目录）
    :param use_iter_files: 如果为 True，则调用 iter_files_with_path，否则调用 iter_download_files
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 迭代器，产生转移结果，有 3 种类型："good"、"fail" 和 "skip"
    """
    from_cid = to_id(from_cid)
    to_pid = to_id(to_pid)
    @as_gen_step
    def upload(attr: dict, pid: int, /):
        @as_gen_step
        def read_range_bytes_or_hash(sign_check: str, /):
            if attr["is_collect"]:
                url = yield from_client.download_url(
                    attr["pickcode"], 
                    use_web_api=True, 
                    async_=async_, 
                    **request_kwargs, 
                )
            else:
                url = yield from_client.download_url(
                    attr["pickcode"], 
                    app="android", 
                    async_=async_, 
                    **request_kwargs, 
                )
            return from_client.request(
                url, 
                headers=dict(url["headers"], range="bytes="+sign_check), 
                parse=False, 
                async_=async_, 
                **request_kwargs, 
            )
        try:
            if not use_iter_files:
                resp = yield from_client.fs_supervision(
                    attr["pickcode"], 
                    async_=async_, 
                    **request_kwargs, 
                )
                check_response(resp)
                info = resp["data"]
                attr.update(
                    id=int(info["file_id"]), 
                    name=info["file_name"], 
                    sha1=info["file_sha1"], 
                    size=int(info["file_size"]), 
                    is_collect=int(info["is_collect"]), 
                    file_type=int(info["file_type"]), 
                )
                if attr["is_collect"] and attr["size"] >= 1024 * 1024 * 115:
                    return {"type": "skip", "attr": attr, "resp": None}
            resp = yield to_client.upload_file_init(
                filename=attr["name"], 
                filesize=attr["size"], 
                filesha1=attr["sha1"], 
                pid=pid, 
                read_range_bytes_or_hash=read_range_bytes_or_hash, 
                base_url=True, 
                async_=async_, 
                **request_kwargs, 
            )
            check_response(resp)
            if resp.get("statuscode"):
                return {"type": "fail", "attr": attr, "resp": resp}
            else:
                return {"type": "good", "attr": attr, "resp": resp}
        except BaseException as e:
            if isinstance(e, OSError) and len(e.args) == 2 and isinstance(e.args[1], dict):
                return {"type": "fail", "attr": attr, "resp": e.args[1], "exc": e}
            else:
                return {"type": "fail", "attr": attr, "resp": None, "exc": e}
    key_of_id = "id" if with_root else "parent_id"
    @as_gen_step
    def get_pid(attr: dict, /):
        if use_iter_files:
            if attr["is_collect"] and attr["size"] >= 1024 * 1024 * 115:
                return Return({"type": "skip", "attr": attr, "resp": None})
            if from_cid:
                dir_ = "/".join(a["name"] for a in dropwhile(
                    lambda a: a[key_of_id] != from_cid, 
                    attr["ancestors"][1:-1], 
                ))
            else:
                dir_ = "/".join(a["name"] for a in attr["ancestors"][1:-1])
        else:
            if from_cid:
                dir_ = "/".join(a["name"] for a in dropwhile(
                    lambda a: a[key_of_id] != from_cid, 
                    attr["dir_ancestors"][1:], 
                ))
            else:
                dir_ = attr["dirname"][1:]
        if dir_ in dir_to_cid:
            return dir_to_cid[dir_]
        else:
            resp = yield to_client.fs_makedirs_app(
                dir_, 
                to_pid, 
                async_=async_, 
                **request_kwargs, 
            )
            check_response(resp)
            pid = dir_to_cid[dir_] = resp["cid"]
            return pid
    dir_to_cid = {"": 0}
    if use_iter_files:
        it = iter_files_with_path(
            from_client, 
            from_cid, 
            normalize_attr=normalize_attr_simple, 
            async_=async_, 
            **request_kwargs, 
        )
    else:
        it = iter_download_files(
            from_client, 
            from_cid, 
            async_=async_, 
            **request_kwargs, 
        )
    if async_:
        return taskgroup_map(upload, it, arg_func=get_pid, max_workers=max_workers)
    else:
        return threadpool_map(upload, it, arg_func=get_pid, max_workers=max_workers)


# TODO: 需要优化，减少代码量
# TODO: 支持一次性把所有文件找完，也支持慢慢处理（可能会风控）
# TODO: 支持 open 接口
@overload
def iter_115_to_115_resume(
    from_client: P115Client, 
    to_client: P115Client, 
    from_cid: int | str = 0, 
    to_pid: int | str = 0, 
    max_workers: int = 8, 
    with_root: bool = True, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator[dict]:
    ...
@overload
def iter_115_to_115_resume(
    from_client: P115Client, 
    to_client: P115Client, 
    from_cid: int | str = 0, 
    to_pid: int | str = 0, 
    max_workers: int = 8, 
    with_root: bool = True, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator[dict]:
    ...
def iter_115_to_115_resume(
    from_client: P115Client, 
    to_client: P115Client, 
    from_cid: int | str = 0, 
    to_pid: int | str = 0, 
    max_workers: int = 8, 
    with_root: bool = True, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> Iterator[dict] | AsyncIterator[dict]:
    """从 115 传到 115（可以跳过已经存在的文件）

    :param from_client: 来源 115 客户端对象
    :param to_client: 去向 115 客户端对象
    :param from_cid: 来源 115 的目录 id 或 pickcode（文件数最好控制在 100 万以内，太多的话，里面多个子文件夹分别传即可）
    :param to_pid: 去向 115 的父目录 id 或 pickcode
    :param max_workers: 最大并发数
    :param with_root: 是否保留 `from_cid` 对应的目录名（如果为 False，则会少 1 级目录）
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 迭代器，产生转移结果，有 3 种类型："good"、"fail" 和 "skip"
    """
    from_cid = to_id(from_cid)
    to_pid = to_id(to_pid)
    @as_gen_step
    def upload(attr: dict, pid: int, /):
        @as_gen_step
        def read_range_bytes_or_hash(sign_check: str, /):
            if attr["is_collect"]:
                url = yield from_client.download_url(
                    attr["pickcode"], 
                    use_web_api=True, 
                    async_=async_, 
                    **request_kwargs, 
                )
            else:
                url = yield from_client.download_url(
                    attr["pickcode"], 
                    app="android", 
                    async_=async_, 
                    **request_kwargs, 
                )
            return from_client.request(
                url, 
                headers=dict(url["headers"], range="bytes="+sign_check), 
                parse=False, 
                async_=async_, 
                **request_kwargs, 
            )
        try:
            resp = yield to_client.upload_file_init(
                filename=attr["name"], 
                filesize=attr["size"], 
                filesha1=attr["sha1"], 
                pid=pid, 
                read_range_bytes_or_hash=read_range_bytes_or_hash, 
                base_url=True, 
                async_=async_, 
                **request_kwargs, 
            )
            check_response(resp)
            if resp.get("statuscode"):
                return {"type": "fail", "attr": attr, "resp": resp}
            else:
                return {"type": "good", "attr": attr, "resp": resp}
        except BaseException as e:
            if isinstance(e, OSError) and len(e.args) == 2 and isinstance(e.args[1], dict):
                return {"type": "fail", "attr": attr, "resp": e.args[1], "exc": e}
            else:
                return {"type": "fail", "attr": attr, "resp": None, "exc": e}
    dirt_to_cid: dict[tuple[str, ...], int] = {}
    key_of_id = "id" if with_root else "parent_id"
    @as_gen_step
    def get_pid(attr: dict, /):
        if attr["is_collect"] and attr["size"] >= 1024 * 1024 * 115:
            return Return({"type": "skip", "attr": attr, "resp": None})
        dirt = tuple(a["name"] for a in dropwhile(
            lambda a: a[key_of_id] != from_cid, 
            attr["ancestors"][1:-1], 
        ))
        try:
            return dirt_to_cid[dirt]
        except KeyError:
            pid = dirt_to_cid[()]
            for i, name in enumerate(dirt, 1):
                p_dirt = dirt[:i]
                if p_dirt in dirt_to_cid:
                    pid = dirt_to_cid[p_dirt]
                else:
                    resp = yield to_client.fs_mkdir(
                        name, 
                        pid, 
                        async_=async_, 
                        **request_kwargs, 
                    )
                    check_response(resp)
                    pid = dirt_to_cid[p_dirt] = int(resp["cid"])
            return pid
    def gen_step():
        from_files: Any = iter_files_with_path(
            from_client, 
            from_cid, 
            normalize_attr=normalize_attr_simple, 
            with_ancestors=True, 
            async_=async_, 
            **request_kwargs, 
        )
        if from_cid:
            resp = yield from_client.fs_file_skim(
                from_cid, 
                async_=async_, 
                **request_kwargs, 
            )
            check_response(resp)
            name = unescape_115_charref(resp["data"][0]["file_name"])
            resp = yield to_client.fs_mkdir(
                name, 
                to_pid, 
                async_=async_, 
                **request_kwargs, 
            )
            if resp.get("errno") == 20004:
                if "/" in name:
                    with with_iter_next(iterdir(
                        to_client, 
                        to_pid, 
                        normalize_attr=normalize_attr_simple, 
                        ensure_file=False, 
                        async_=async_, 
                        **request_kwargs, 
                    )) as get_next:
                        while True:
                            attr = yield get_next()
                            if attr["name"] == name:
                                to_cid = attr["id"]
                                break
                else:
                    resp = yield to_client.fs_makedirs_app(
                        name, 
                        to_pid, 
                        async_=async_, 
                        **request_kwargs, 
                    )
                    check_response(resp)
                    to_cid = int(resp["cid"])
                dirt_to_cid[()] = to_cid
                id_to_dirnode: dict[int, tuple[str, int]] = {}
                to_files: Any = iter_files_with_path(
                    to_client, 
                    to_cid, 
                    id_to_dirnode=id_to_dirnode, 
                    normalize_attr=normalize_attr_simple, 
                    with_ancestors=True, 
                    async_=async_, 
                    **request_kwargs, 
                )
                if async_:
                    from_files, to_files = yield to_list(
                        taskgroup_map(to_list, (from_files, to_files), max_workers=2))
                else:
                    from_files, to_files = threadpool_map(list, (from_files, to_files), max_workers=2)
                while to_cid:
                    _, to_cid = id_to_dirnode.pop(to_cid)
                cid_to_dirt: dict[int, tuple[str, ...]] = {}
                def get_dirt(cid: int, /) -> tuple[str, ...]:
                    if cid not in id_to_dirnode:
                        return ()
                    name, pid = id_to_dirnode[cid]
                    if pid in cid_to_dirt:
                        p_dirt = cid_to_dirt[pid]
                    else:
                        p_dirt = get_dirt(pid)
                    dirt = (*p_dirt, name)
                    dirt_to_cid[dirt] = cid
                    cid_to_dirt[cid] = dirt
                    return dirt
                for cid, (name, pid) in id_to_dirnode.items():
                    if cid not in cid_to_dirt:
                        get_dirt(cid)
                del cid_to_dirt, id_to_dirnode
                to_cid = dirt_to_cid[()]
                seen = {
                    tuple(a["name"] for a in dropwhile(
                        lambda a: a["parent_id"] != to_cid, 
                        attr["ancestors"][1:], 
                    )) for attr in to_files
                }
                from_files = [
                    attr for attr in from_files
                    if tuple(a["name"] for a in dropwhile(
                        lambda a: a["parent_id"] != from_cid, 
                        attr["ancestors"][1:], 
                    )) not in seen
                ]
                del to_files, seen
            else:
                check_response(resp)
                dirt_to_cid[()] = int(resp["cid"])
        else:
            dirt_to_cid[()] = 0
        if async_:
            return YieldFrom(taskgroup_map(
                upload, 
                from_files, 
                arg_func=get_pid, 
                max_workers=max_workers, 
            ))
        else:
            return YieldFrom(threadpool_map(
                upload, 
                from_files, 
                arg_func=get_pid, 
                max_workers=max_workers, 
            ))
    return run_gen_step_iter(gen_step, async_)


@overload
def upload_init(
    client: str | PathLike | P115Client | P115OpenClient, 
    file: Buffer | str | PathLike | URL | SupportsGeturl | SupportsRead, 
    pid: int | str = 0, 
    filename: str = "", 
    filesha1: str = "", 
    filesize: int = -1, 
    endpoint: str = "http://oss-cn-shenzhen.aliyuncs.com", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def upload_init(
    client: str | PathLike | P115Client | P115OpenClient, 
    file: Buffer | str | PathLike | URL | SupportsGeturl | SupportsRead, 
    pid: int | str = 0, 
    filename: str = "", 
    filesha1: str = "", 
    filesize: int = -1, 
    endpoint: str = "http://oss-cn-shenzhen.aliyuncs.com", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def upload_init(
    client: str | PathLike | P115Client | P115OpenClient, 
    file: Buffer | str | PathLike | URL | SupportsGeturl | SupportsRead, 
    pid: int | str = 0, 
    filename: str = "", 
    filesha1: str = "", 
    filesize: int = -1, 
    endpoint: str = "http://oss-cn-shenzhen.aliyuncs.com", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """准备上传，获取必要信息，可能秒传成功

    :param client: 115 客户端或 cookies
    :param file: 待上传的文件或其路径
    :param pid: 上传文件到目录的 id 或 pickcode
    :param filename: 文件名，若为空则自动确定
    :param filesha1: 文件的 sha1 哈希值，若为空则自动计算
    :param filesize: 文件大小，若为负数则自动计算
    :param endpoint: 上传目的网址
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 响应信息，如果有字段 "reuse" 为 True，则说明秒传成功
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if isinstance(client, P115Client):
        return oss_upload_init(
            file=file, 
            pid=to_id(pid), 
            filename=filename, 
            filesha1=filesha1, 
            filesize=filesize, 
            user_id=client.user_id, 
            user_key=client.user_key, 
            endpoint=endpoint, 
            async_=async_, 
            **request_kwargs, 
        )
    else:
        request_kwargs["headers"] = dict(
            request_kwargs.get("headers") or (), 
            authorization=client.headers["authorization"], 
        )
        return oss_upload_init(
            file=file, 
            pid=to_id(pid), 
            filename=filename, 
            filesha1=filesha1, 
            filesize=filesize, 
            endpoint=endpoint, 
            async_=async_, 
            **request_kwargs, 
        )


class P115MultipartUpload:
    """待分块上传对象

    :param url: HTTP 请求链接，包含存储桶和对象的名字
    :param path: 待上传的文件路径
    :param callback: 回调数据
    :param upload_id: 上传任务 id

    下面是一个上传的例子，会在命令行显示进度条

    .. code::

        from pathlib import Path
        from p115client import P115Client
        from p115client.tool import P115MultipartUpload

        client = P115Client(Path("~/115-cookies.txt").expanduser())

        # NOTE: 待上传文件的路径（同样也支持 URL）
        path = "/path/to/file"

        uploader = P115MultipartUpload.from_path(path, user_id=client.user_id, user_key=client.user_key)
        # NOTE: 返回字典说明秒传成功
        if isinstance(uploader, dict):
            print(uploader)
        else:
            from os.path import getsize
            # NOTE: 你可以随意指定其它各种进度条模块，或者自己写的函数
            from tqdm import tqdm

            # NOTE: 文件总大小需要你自己获取，`reporthook`只做增量推送
            with tqdm(total=getsize(path), unit="B", unit_scale=True, desc="Uploading") as t:
                # NOTE: `iter_upload` 支持其它请求模块，例如 urllib3
                #     from urllib3_request import request
                #     uploader.iter_upload(request=request)
                for _ in uploader.iter_upload(reporthook=t.update):
                    pass
            print(uploader.complete())

    你也可以自己写一个进度条

    .. code::

        from collections import deque
        from time import perf_counter

        def make_reporthook(total: None | int = None):
            dq: deque[tuple[int, float]] = deque(maxlen=64)
            push = dq.append
            read_num = 0
            push((read_num, perf_counter()))
            while True:
                read_num += yield
                cur_t = perf_counter()
                speed = (read_num - dq[0][0]) / 1024 / 1024 / (cur_t - dq[0][1])
                if total:
                    percentage = read_num / total * 100
                    print(f"\\r\\x1b[K{read_num} / {total} | {speed:.2f} MB/s | {percentage:.2f} %", end="", flush=True)
                else:
                    print(f"\\r\\x1b[K{read_num} | {speed:.2f} MB/s", end="", flush=True)
                push((read_num, cur_t))

    然后像下面这样使用

    .. code::

        for _ in uploader.iter_upload(reporthook=make_reporthook(getsize(path)).send):
            pass
    """
    __slots__ = ("url", "path", "callback", "upload_id", "_result")

    def __init__(
        self, 
        /, 
        url: str, 
        path: str | PathLike | URL | SupportsGeturl, 
        callback: dict, 
        upload_id: str = "", 
    ):
        self.url = url
        if isinstance(path, PathLike):
            path = fsdecode(path)
        elif isinstance(path, URL):
            path = str(path)
        elif isinstance(path, SupportsGeturl):
            path = path.geturl()
        self.path = cast(str, path)
        self.callback = callback
        if not upload_id:
            upload_id = oss_multipart_upload_init(url)
        self.upload_id = upload_id
        self._result = None

    def __repr__(self, /) -> str:
        cls = type(self)
        mod = cls.__module__
        name = cls.__qualname__
        url = self.url
        path = self.path
        callback = self.callback
        upload_id = self.upload_id
        return f"{mod}.{name}({url=!r}, {path=!r}, {callback=!r}, {upload_id=!r})"

    @overload
    @classmethod
    def from_path(
        cls, 
        /, 
        path: str | PathLike | URL | SupportsGeturl, 
        pid: int | str = 0, 
        filename: str = "", 
        filesha1: str = "", 
        filesize: int = -1, 
        user_id: int | str = "", 
        user_key: str = "", 
        endpoint: str = "http://oss-cn-shenzhen.aliyuncs.com", 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> dict | P115MultipartUpload:
        ...
    @overload
    @classmethod
    def from_path(
        cls, 
        /, 
        path: str | PathLike | URL | SupportsGeturl, 
        pid: int | str = 0, 
        filename: str = "", 
        filesha1: str = "", 
        filesize: int = -1, 
        user_id: int | str = "", 
        user_key: str = "", 
        endpoint: str = "http://oss-cn-shenzhen.aliyuncs.com", 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, dict | P115MultipartUpload]:
        ...
    @classmethod
    def from_path(
        cls, 
        /, 
        path: str | PathLike | URL | SupportsGeturl, 
        pid: int | str = 0, 
        filename: str = "", 
        filesha1: str = "", 
        filesize: int = -1, 
        user_id: int | str = "", 
        user_key: str = "", 
        endpoint: str = "http://oss-cn-shenzhen.aliyuncs.com", 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> dict | P115MultipartUpload | Coroutine[Any, Any, dict | P115MultipartUpload]:
        """准备上传，获取必要信息，可能秒传成功

        :param path: 待上传的文件路径
        :param pid: 上传文件到目录的 id 或 pickcode
        :param filename: 文件名，若为空则自动确定
        :param filesha1: 文件的 sha1 哈希值，若为空则自动计算
        :param filesize: 文件大小，若为负数则自动计算
        :param user_id: 用户 id
        :param user_key: 用户的 key
        :param endpoint: 上传目的网址
        :param async_: 是否异步
        :param request_kwargs: 其它请求参数

        :return: 秒传成功的响应或者待分块上传对象
        """
        def gen_step():
            if user_id and user_key:
                resp = yield oss_upload_init(
                    file=path, 
                    pid=to_id(pid), 
                    filename=filename, 
                    filesha1=filesha1, 
                    filesize=filesize, 
                    user_id=user_id, 
                    user_key=user_key, 
                    endpoint=endpoint, 
                    async_=async_, 
                    **request_kwargs, 
                )
            else:
                try:
                    headers = request_kwargs["headers"]
                except KeyError as e:
                    raise TypeError(f"{cls.from_path!r} missing 1 required keyword-only argument: 'headers'") from e
                headers = request_kwargs["headers"] = dict_map(headers or (), key=str.lower)
                if "authorization" in headers:
                    resp = yield oss_upload_init(
                        file=path, 
                        pid=to_id(pid), 
                        filename=filename, 
                        filesha1=filesha1, 
                        filesize=filesize, 
                        endpoint=endpoint, 
                        async_=async_, 
                        **request_kwargs, 
                    )
                elif "cookie" in headers:
                    client = P115Client(headers["cookie"])
                    resp = yield oss_upload_init(
                        file=path, 
                        pid=to_id(pid), 
                        filename=filename, 
                        filesha1=filesha1, 
                        filesize=filesize, 
                        user_id=client.user_id, 
                        user_key=client.user_key, 
                        endpoint=endpoint, 
                        async_=async_, 
                        **request_kwargs, 
                    )
                else:
                    raise ValueError("please provide the request header: 'authorization' or 'cookie'")
            check_response(resp)
            if resp["reuse"]:
                return resp
            data = resp["data"]
            url = data["url"]
            upload_id = yield oss_multipart_upload_init(
                url, 
                async_=async_, 
                **request_kwargs, 
            )
            return cls(url, path, data["callback"], upload_id)
        return run_gen_step(gen_step, async_)

    @property
    def completed(self, /) -> bool:
        "是否已完成"
        return bool(self._result)

    @property
    def result(self, /) -> None | dict:
        "完成后的结果"
        return self._result

    @property
    def succeeded(self, /) -> bool:
        "结果是否成功"
        return bool(self._result and self._result["state"])

    @overload
    def complete(
        self, 
        /, 
        parts: None | list[dict] = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> dict:
        ...
    @overload
    def complete(
        self, 
        /, 
        parts: None | list[dict] = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, dict]:
        ...
    def complete(
        self, 
        /, 
        parts: None | list[dict] = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> dict | Coroutine[Any, Any, dict]:
        """完成分块上传

        :param parts: 分块信息列表
        :param async_: 是否异步
        :param request_kwargs: 其它请求参数

        :return: 接口响应
        """
        if self.completed:
            raise RuntimeError("already completed")
        def gen_step():
            nonlocal parts
            if parts is None:
                parts = yield self.list_parts(async_=async_, **request_kwargs)
            resp = self._result = yield oss_multipart_upload_complete(
                self.url, 
                self.callback, 
                self.upload_id, 
                parts=parts, 
                async_=async_, 
                **request_kwargs, 
            )
            return resp
        return run_gen_step(gen_step, async_)

    @overload
    def list_parts(
        self, 
        /, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> list[dict]:
        ...
    @overload
    def list_parts(
        self, 
        /, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, list[dict]]:
        ...
    def list_parts(
        self, 
        /, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> list[dict] | Coroutine[Any, Any, list[dict]]:
        """罗列已上传的分块信息

        :param async_: 是否异步
        :param request_kwargs: 其它请求参数

        :return: 分块信息列表
        """
        def gen_step():
            parts: list[dict] = []
            yield foreach(
                parts.append, 
                oss_multipart_part_iter(
                    self.url, 
                    self.upload_id, 
                    async_=async_, 
                    **request_kwargs, 
                ), 
            )
            return parts
        return run_gen_step(gen_step, async_)

    @overload
    def upload_url(
        self, 
        /, 
        part_number: int = 1, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> tuple[str, dict]:
        ...
    @overload
    def upload_url(
        self, 
        /, 
        part_number: int = 1, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> Coroutine[Any, Any, tuple[str, dict]]:
        ...
    def upload_url(
        self, 
        /, 
        part_number: int = 1, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> tuple[str, dict] | Coroutine[Any, Any, tuple[str, dict]]:
        """获取分块上传的链接和请求头

        .. caution::
            这个接口只用来获取上传链接和请求头，并不会做实际的上传

        :param part_number: 分块编号（从 1 开始）
        :param async_: 是否异步
        :param request_kwargs: 其它请求参数

        :return: 上传链接 和 请求头 的 2 元组
        """
        return oss_multipart_upload_url(
            self.url, 
            self.upload_id, 
            part_number, 
            async_=async_, 
            **request_kwargs, 
        )

    @overload
    def iter_upload(
        self, 
        /, 
        partsize: int = 10485760, 
        reporthook: None | Callable[[int], Any] = None, 
        opener: None | Callable[[str, int], SupportsRead | Iterable[Buffer]] = None, 
        *, 
        async_: Literal[False] = False, 
        **request_kwargs, 
    ) -> Iterator[dict]:
        ...
    @overload
    def iter_upload(
        self, 
        /, 
        partsize: int = 10485760, 
        reporthook: None | Callable[[int], Any] = None, 
        opener: None | Callable[[str, int], SupportsRead | Awaitable[SupportsRead] | Iterable[Buffer] | AsyncIterable[Buffer]] = None, 
        *, 
        async_: Literal[True], 
        **request_kwargs, 
    ) -> AsyncIterator[dict]:
        ...
    def iter_upload(
        self, 
        /, 
        partsize: int = 10485760, 
        reporthook: None | Callable[[int], Any] = None, 
        opener: (
            None | 
            Callable[[str, int], SupportsRead | Iterable[Buffer]] | 
            Callable[[str, int], SupportsRead | Awaitable[SupportsRead] | Iterable[Buffer] | AsyncIterable[Buffer]]
        ) = None, 
        *, 
        async_: Literal[False, True] = False, 
        **request_kwargs, 
    ) -> Iterator[dict] | AsyncIterator[dict]:
        """逐个上传分块

        .. attention::
            上传完成后，并不会提交，请手动调用 ``.complete()`` 方法

        .. note::
            - 可随意搭配请求模块: 指定 ``request`` 参数
            - 可随意搭配进度条: 指定 ``reporthook`` 参数
            - 可随意搭配文件打开器: 指定 ``opener`` 参数

        .. note::
            如果想把把上传过程外包出去，由其它任何工具完成，则调用 ``upload_url(part_number)`` 方法获得指定分块的上传链接和带签名请求头

            或者调用 ``iter_upload_url(part_number_start)`` 创建一个迭代器，从某个分块编号开始，获得一系列的上传链接和带签名请求头

        :param partsize: 分块大小
        :param reporthook: 回调函数，可以用来统计已上传的数据量或者展示进度条
        :param opener: 打开文件路径（本地路径或 URL）并从指定位置开始，如果为 None，则用默认方式
        :param async_: 是否异步
        :param request_kwargs: 其它请求参数

        :return: 迭代器，产生各个刚上传完成的分块信息
        """
        def gen_step():
            parts = yield self.list_parts(async_=async_, **request_kwargs)
            skipsize = sum(p["Size"] for p in parts)
            if reporthook is not None:
                reporthook(skipsize)
            path = self.path
            if opener is not None:
                file = opener(path, skipsize)
                if async_ and isawaitable(file):
                    file = yield file
            elif path.startswith(("http://", "https://")):
                if async_:
                    from httpfile import AsyncHTTPFileReader
                    async def process():
                        return await AsyncHTTPFileReader.new(path, start=skipsize)
                    file = yield process()
                else:
                    from httpfile import HTTPFileReader
                    file = HTTPFileReader(path, start=skipsize)
            else:
                file = open(path, "rb")
                if skipsize:
                    file.seek(skipsize)
            file = cast(SupportsRead | Iterable[Buffer] | AsyncIterable[Buffer], file)
            try:
                yield YieldFrom(oss_multipart_upload_part_iter(
                    url=self.url, 
                    file=file, # type: ignore
                    upload_id=self.upload_id, 
                    partsize=partsize, 
                    part_number_start=len(parts)+1, 
                    reporthook=reporthook, 
                    async_=async_, # type: ignore
                    **request_kwargs, 
                ))
            finally:
                if async_:
                    if hasattr(file, "aclose"):
                        yield file.aclose()
                    elif hasattr(file, "close"):
                        ret = file.close()
                        if isawaitable(ret):
                            yield ret
                elif hasattr(file, "close"):
                    file.close()
        return run_gen_step_iter(gen_step, async_)

    def iter_upload_url(
        self, 
        /, 
        part_number_start: int = 1, 
        headers: None | dict[str, str] = None, 
    ) -> Iterator[tuple[str, dict]]:
        """逐个获取上传链接和请求头

        .. caution::
            这个接口只用来获取上传链接和请求头，并不会做实际的上传，而且也不会判断总共有多少个分块，而是无限生成

        :param part_number_start: 开始的分块编号，从 1 开始
        :param headers: 默认的请求头，会被扩展

        :return: 迭代器，产生上传链接和请求头（带签名）
        """
        if part_number_start <= 0:
            part_number_start = 1
        get_url = self.upload_url
        for part_number in count(part_number_start):
            yield get_url(part_number, headers=headers)

# TODO: 增加一个工具函数，用于从某个本地目录下载到网盘目录，允许提供自定义的进度条调用
# TODO: 增加一个工具函数，用于在两个115网盘之间的转移，允许提供自定义的进度条调用



#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = [
    "iter_fs_files", "iter_fs_files_serialized", "iter_fs_files_threaded", 
    "iter_fs_files_asynchronized", 
]
__doc__ = "这个模块利用 P115Client.fs_files 方法做了一些封装"

from asyncio import (
    shield, sleep as async_sleep, wait_for, 
    Semaphore as AsyncSemaphore, Task, TaskGroup, 
)
from collections import deque
from collections.abc import AsyncIterator, Awaitable, Callable, Iterator
from concurrent.futures import Future, ThreadPoolExecutor
from copy import copy
from inspect import isawaitable
from itertools import cycle
from os import PathLike
from time import sleep, time
from typing import overload, Any, Final, Literal
from warnings import warn

from errno2 import errno
from http_response import get_status_code, is_timeouterror
from iterutils import run_gen_step_iter, Yield

from ..client import check_response, P115Client, P115OpenClient
from ..exception import throw, P115DataError, P115Warning


get_webapi_origin: Final = cycle(("http://web.api.115.com", "https://webapi.115.com")).__next__
get_proapi_origin: Final = cycle(("http://pro.api.115.com", "https://proapi.115.com")).__next__


@overload
def iter_fs_files(
    client: str | PathLike | P115Client | P115OpenClient, 
    payload: int | str | dict = 0, 
    /, 
    page_size: int = 7_000, 
    first_page_size: int = 0, 
    count: int = -1, 
    callback: None | Callable[[dict], Any] = None, 
    app: str = "web", 
    raise_for_changed_count: bool = False, 
    cooldown: None | float = None, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator[dict]:
    ...
@overload
def iter_fs_files(
    client: str | PathLike | P115Client | P115OpenClient, 
    payload: int | str | dict = 0, 
    /, 
    page_size: int = 7_000, 
    first_page_size: int = 0, 
    count: int = -1, 
    callback: None | Callable[[dict], Any] = None, 
    app: str = "web", 
    raise_for_changed_count: bool = False, 
    cooldown: None | float = None, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator[dict]:
    ...
def iter_fs_files(
    client: str | PathLike | P115Client | P115OpenClient, 
    payload: int | str | dict = 0, 
    /, 
    page_size: int = 7_000, 
    first_page_size: int = 0, 
    count: int = -1, 
    callback: None | Callable[[dict], Any] = None, 
    app: str = "web", 
    raise_for_changed_count: bool = False, 
    cooldown: None | float = None, 
    max_workers: None | int = None, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> AsyncIterator[dict] | Iterator[dict]:
    """拉取一个目录中的文件或目录的数据

    :param client: 115 网盘客户端对象
    :param payload: 目录的 id 或者详细的查询参数
    :param page_size: 分页大小，如果 <= 0，则自动确定
    :param first_page_size: 首次拉取的分页大小，如果 <= 0，则自动确定
    :param count: 文件总数
    :param callback: 回调函数，调用后，会获得一个值，会添加到返回值中，key 为 "callback"
    :param app: 使用此设备的接口
    :param raise_for_changed_count: 分批拉取时，发现总数发生变化后，是否报错
    :param cooldown: 冷却时间，单位为秒。如果为 None，则用默认值（非并发时为 0，并发时为 1）
    :param max_workers: 最大并发数，如果为 None 或 < 0 则自动确定，如果为 0 则单工作者惰性执行
    :param async_: 是否异步
    :param request_kwargs: 其它 http 请求参数，会传给具体的请求函数，默认的是 httpx，可用参数 request 进行设置

    :return: 迭代器，每次返回一次接口调用的结果
    """
    if max_workers == 0:
        request_kwargs["async_"] = async_
        method: Callable = iter_fs_files_serialized
    else:
        request_kwargs["max_workers"] = max_workers
        if async_:
            method = iter_fs_files_asynchronized
        else:
            method = iter_fs_files_threaded
    if cooldown is not None:
        request_kwargs["cooldown"] = cooldown
    return method(
        client, 
        payload, 
        page_size=page_size, 
        first_page_size=first_page_size, 
        count=count, 
        callback=callback, 
        app=app, 
        raise_for_changed_count=raise_for_changed_count, 
        **request_kwargs, 
    )


@overload
def iter_fs_files_serialized(
    client: str | PathLike | P115Client | P115OpenClient, 
    payload: int | str | dict = 0, 
    /, 
    page_size: int = 7_000, 
    first_page_size: int = 0, 
    count: int = -1, 
    callback: None | Callable[[dict], Any] = None, 
    app: str = "web", 
    raise_for_changed_count: bool = False, 
    cooldown: float = 0, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> Iterator[dict]:
    ...
@overload
def iter_fs_files_serialized(
    client: str | PathLike | P115Client | P115OpenClient, 
    payload: int | str | dict = 0, 
    /, 
    page_size: int = 7_000, 
    first_page_size: int = 0, 
    count: int = -1, 
    callback: None | Callable[[dict], Any] = None, 
    app: str = "web", 
    raise_for_changed_count: bool = False, 
    cooldown: float = 0, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> AsyncIterator[dict]:
    ...
def iter_fs_files_serialized(
    client: str | PathLike | P115Client | P115OpenClient, 
    payload: int | str | dict = 0, 
    /, 
    page_size: int = 7_000, 
    first_page_size: int = 0, 
    count: int = -1, 
    callback: None | Callable[[dict], Any] = None, 
    app: str = "web", 
    raise_for_changed_count: bool = False, 
    cooldown: float = 0, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> AsyncIterator[dict] | Iterator[dict]:
    """拉取一个目录中的文件或目录的数据

    :param client: 115 网盘客户端对象
    :param payload: 目录的 id 或者详细的查询参数
    :param page_size: 分页大小，如果 <= 0，则自动确定
    :param first_page_size: 首次拉取的分页大小，如果 <= 0，则自动确定
    :param count: 文件总数
    :param callback: 回调函数，调用后，会获得一个值，会添加到返回值中，key 为 "callback"
    :param app: 使用此设备的接口
    :param raise_for_changed_count: 分批拉取时，发现总数发生变化后，是否报错
    :param cooldown: 冷却时间，单位为秒
    :param async_: 是否异步
    :param request_kwargs: 其它 http 请求参数，会传给具体的请求函数，默认的是 httpx，可用参数 request 进行设置

    :return: 迭代器，每次返回一次接口调用的结果
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if page_size <= 0:
        page_size = 7_000
    fs_files: Callable
    if not isinstance(client, P115Client) or app == "open":
        page_size = min(page_size, 1150)
        fs_files = client.fs_files_open
    elif app in ("", "web", "desktop", "harmony"):
        page_size = min(page_size, 1150)
        request_kwargs.setdefault("base_url", get_webapi_origin)
        fs_files = client.fs_files
    elif app == "aps":
        page_size = min(page_size, 1200)
        fs_files = client.fs_files_aps
    else:
        request_kwargs.setdefault("base_url", get_proapi_origin)
        request_kwargs["app"] = app
        fs_files = client.fs_files_app
    if first_page_size <= 0:
        first_page_size = page_size
    if isinstance(payload, (int, str)):
        payload = {"cid": client.to_id(payload)}
    payload = {
        "asc": 1, "cid": 0, "fc_mix": 1, "o": "user_ptime", "offset": 0, 
        "limit": first_page_size, "show_dir": 1, **payload, 
    }
    cid = int(payload["cid"])
    def gen_step():
        nonlocal count, first_page_size
        last_call_ts: float = 0
        while True:
            while True:
                try:
                    if cooldown > 0 and (delta := last_call_ts + cooldown - time()) > 0:
                        if async_:
                            yield async_sleep(delta)
                        else:
                            sleep(delta)
                        last_call_ts = time()
                    resp = yield fs_files(payload, async_=async_, **request_kwargs)
                    check_response(resp)
                except P115DataError:
                    if payload["limit"] <= 1150:
                        raise
                    payload["limit"] -= 1_000
                    if payload["limit"] < 1150:
                        payload["limit"] = 1150
                    continue
                if cid and int(resp["path"][-1]["cid"]) != cid:
                    if count < 0:
                        throw(errno.ENOTDIR, cid)
                    else:
                        throw(errno.ENOENT, cid)
                count_new = int(resp["count"])
                if count < 0:
                    count = count_new
                elif count != count_new:
                    message = f"cid={cid} detected count changes during iteration: {count} -> {count_new}"
                    if raise_for_changed_count:
                        throw(errno.EBUSY, message)
                    else:
                        warn(message, category=P115Warning)
                    count = count_new
                if callback is not None:
                    resp["callback"] = yield callback(resp)
                break
            yield Yield(resp)
            payload["offset"] += len(resp["data"])
            if payload["offset"] >= count:
                break
            if first_page_size != page_size:
                payload["limit"] = page_size
                first_page_size = page_size
    return run_gen_step_iter(gen_step, async_)


def iter_fs_files_threaded(
    client: str | PathLike | P115Client | P115OpenClient, 
    payload: int | str | dict = 0, 
    /, 
    page_size: int = 7_000, 
    first_page_size: int = 0, 
    count: int = -1, 
    callback: None | Callable[[dict], Any] = None, 
    app: str = "web", 
    raise_for_changed_count: bool = False, 
    cooldown: float = 1, 
    max_workers: None | int = None, 
    **request_kwargs, 
) -> Iterator[dict]:
    """多线程并发拉取一个目录中的文件或目录的数据

    :param client: 115 网盘客户端对象
    :param payload: 目录的 id、pickcode 或者详细的查询参数
    :param page_size: 分页大小，如果 <= 0，则自动确定
    :param first_page_size: 第 1 次拉取的分页大小，如果指定此参数，则会等待这次请求返回，才会开始后续，也即非并发
    :param count: 文件总数
    :param callback: 回调函数，调用后，会获得一个值，会添加到返回值中，key 为 "callback"
    :param app: 使用此设备的接口
    :param raise_for_changed_count: 分批拉取时，发现总数发生变化后，是否报错
    :param cooldown: 冷却时间，单位为秒
    :param max_workers: 最大工作线程数，如果为 None，则自动确定
    :param request_kwargs: 其它 http 请求参数，会传给具体的请求函数，默认的是 httpx，可用参数 request 进行设置

    :return: 迭代器
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if page_size <= 0:
        page_size = 7_000
    fs_files: Callable[..., Awaitable[dict]]
    if not isinstance(client, P115Client) or app == "open":
        page_size = min(page_size, 1150)
        fs_files = client.fs_files_open
    elif app in ("", "web", "desktop", "harmony"):
        page_size = min(page_size, 1150)
        request_kwargs.setdefault("base_url", get_webapi_origin)
        fs_files = client.fs_files
    elif app == "aps":
        page_size = min(page_size, 1200)
        fs_files = client.fs_files_aps
    else:
        request_kwargs.setdefault("base_url", get_proapi_origin)
        request_kwargs["app"] = app
        fs_files = client.fs_files_app
    if isinstance(payload, (int, str)):
        payload = {"cid": client.to_id(payload)}
    if first_page_size <= 0:
        first_page_size = page_size
    payload = {
        "asc": 1, "cid": 0, "fc_mix": 1, "o": "user_ptime", "offset": 0, 
        "limit": first_page_size, "show_dir": 1, **payload, 
    }
    cid = int(payload["cid"])
    def get_files(payload: dict, /):
        nonlocal count
        resp = fs_files(payload, **request_kwargs)
        check_response(resp)
        if cid and int(resp["path"][-1]["cid"]) != cid:
            if count < 0:
                throw(errno.ENOTDIR, cid)
            else:
                throw(errno.ENOENT, cid)
        count_new = int(resp["count"])
        if count < 0:
            count = count_new
        elif count != count_new:
            message = f"cid={cid} detected count changes during iteration: {count} -> {count_new}"
            if raise_for_changed_count:
                throw(errno.EBUSY, message)
            else:
                warn(message, category=P115Warning)
            count = count_new
        if callback is not None:
            resp["callback"] = callback(resp)
        return resp
    dq: deque[tuple[Future, int]] = deque()
    push, pop = dq.append, dq.popleft
    if max_workers is not None and max_workers <= 0:
        max_workers = None
    executor = ThreadPoolExecutor(max_workers=max_workers)
    submit = executor.submit
    last_call_ts: int | float = 0
    def make_future(args: None | dict = None, /) -> Future:
        nonlocal last_call_ts
        if args is None:
            args = copy(payload)
        last_call_ts = time()
        return submit(get_files, args)
    try:
        future = make_future()
        payload["limit"] = page_size
        offset = payload["offset"]
        while True:
            try:
                if first_page_size == page_size:
                    resp = future.result(max(0, last_call_ts + cooldown - time()))
                else:
                    resp = future.result()
                    first_page_size = page_size
            except TimeoutError:
                payload["offset"] += page_size
                if count < 0 or payload["offset"] < count:
                    push((make_future(), payload["offset"]))
            except BaseException as e:
                if get_status_code(e) >= 400 or not is_timeouterror(e):
                    raise
                future = make_future({**payload, "offset": offset})
            else:
                yield resp
                reach_end = offset != resp["offset"] or count >= 0 and offset + len(resp["data"]) >= count
                will_continue = False
                while dq:
                    future, offset = pop()
                    if count < 0 or offset < count:
                        will_continue = True
                        break
                    future.cancel()
                if will_continue:
                    continue
                elif reach_end:
                    break
                else:
                    offset = payload["offset"] = payload["offset"] + page_size
                    if count >= 0 and offset >= count:
                        break
                    future = make_future()
    finally:
        executor.shutdown(False, cancel_futures=True)


async def iter_fs_files_asynchronized(
    client: str | PathLike | P115Client | P115OpenClient, 
    payload: int | str | dict = 0, 
    /, 
    page_size: int = 7_000, 
    first_page_size: int = 0, 
    count: int = -1, 
    callback: None | Callable[[dict], Any] = None, 
    app: str = "web", 
    raise_for_changed_count: bool = False, 
    cooldown: float = 1, 
    max_workers: None | int = None, 
    **request_kwargs, 
) -> AsyncIterator[dict]:
    """异步并发拉取一个目录中的文件或目录的数据

    :param client: 115 网盘客户端对象
    :param payload: 目录的 id、pickcode 或者详细的查询参数
    :param page_size: 分页大小，如果 <= 0，则自动确定
    :param first_page_size: 第 1 次拉取的分页大小，如果指定此参数，则会等待这次请求返回，才会开始后续，也即非并发
    :param count: 文件总数
    :param callback: 回调函数，调用后，会获得一个值，会添加到返回值中，key 为 "callback"
    :param app: 使用此设备的接口
    :param raise_for_changed_count: 分批拉取时，发现总数发生变化后，是否报错
    :param cooldown: 冷却时间，单位为秒
    :param max_workers: 最大工作协程数，如果为 None 或 <= 0，则为 64
    :param request_kwargs: 其它 http 请求参数，会传给具体的请求函数，默认的是 httpx，可用参数 request 进行设置

    :return: 异步迭代器
    """
    if isinstance(client, (str, PathLike)):
        client = P115Client(client, check_for_relogin=True)
    if page_size <= 0:
        page_size = 7_000
    fs_files: Callable[..., Awaitable[dict]]
    if not isinstance(client, P115Client) or app == "open":
        page_size = min(page_size, 1150)
        fs_files = client.fs_files_open
    elif app in ("", "web", "desktop", "harmony"):
        page_size = min(page_size, 1150)
        request_kwargs.setdefault("base_url", get_webapi_origin)
        fs_files = client.fs_files
    elif app == "aps":
        page_size = min(page_size, 1200)
        fs_files = client.fs_files_aps
    else:
        request_kwargs.setdefault("base_url", get_proapi_origin)
        request_kwargs["app"] = app
        fs_files = client.fs_files_app
    if first_page_size <= 0:
        first_page_size = page_size
    if isinstance(payload, (int, str)):
        payload = {"cid": client.to_id(payload)}
    payload = {
        "asc": 1, "cid": 0, "fc_mix": 1, "o": "user_ptime", "offset": 0, 
        "limit": first_page_size, "show_dir": 1, **payload, 
    }
    cid = int(payload["cid"])
    if max_workers is None or max_workers <= 0:
        max_workers = 64
    sema = AsyncSemaphore(max_workers)
    async def get_files(payload: dict, /):
        nonlocal count
        async with sema:
            resp = await fs_files(payload, async_=True, **request_kwargs)
        check_response(resp)
        if cid and int(resp["path"][-1]["cid"]) != cid:
            if count < 0:
                throw(errno.ENOTDIR, cid)
            else:
                throw(errno.ENOENT, cid)
        count_new = int(resp["count"])
        if count < 0:
            count = count_new
        elif count != count_new:
            message = f"cid={cid} detected count changes during iteration: {count} -> {count_new}"
            if raise_for_changed_count:
                throw(errno.EBUSY, message)
            else:
                warn(message, category=P115Warning)
            count = count_new
        if callback is not None:
            ret = callback(resp)
            if isawaitable(ret):
                ret = await ret
            resp["callback"] = ret
        return resp
    dq: deque[tuple[Task, int]] = deque()
    push, pop = dq.append, dq.popleft
    async with TaskGroup() as tg:
        create_task = tg.create_task
        last_call_ts: float = 0
        def make_task(args: None | dict = None, /) -> Task:
            nonlocal last_call_ts
            if args is None:
                args = copy(payload)
            last_call_ts = time()
            return create_task(get_files(args))
        task   = make_task()
        payload["limit"] = page_size
        offset = payload["offset"]
        while True:
            try:
                if first_page_size == page_size:
                    resp = await wait_for(shield(task), max(0, last_call_ts + cooldown - time()))
                else:
                    resp = await task
                    first_page_size = page_size
            except TimeoutError:
                payload["offset"] += page_size
                if count < 0 or payload["offset"] < count:
                    push((make_task(), payload["offset"]))
            except BaseException as e:
                if get_status_code(e) >= 400 or not is_timeouterror(e):
                    raise
                task = make_task({**payload, "offset": offset})
            else:
                yield resp
                reach_end = offset != resp["offset"] or count >= 0 and offset + len(resp["data"]) >= count
                will_continue = False
                while dq:
                    task, offset = pop()
                    if count < 0 or offset < count:
                        will_continue = True
                        break
                    task.cancel()
                if will_continue:
                    continue
                elif reach_end:
                    break
                else:
                    offset = payload["offset"] = payload["offset"] + page_size
                    if count >= 0 and offset >= count:
                        break
                    task = make_task()

# TODO: 这个模块对外暴露，应该只有一个 iter_fs_files，其它接口，fs_files 部分由外部传入
# TODO: 移除 callback 参数
